/**************************************************************************
* DSemu - The Next Generation                                             *
* Portable ARM7TDMI core: Plugin implementation [arm7tdmi.cpp]            *
* Copyright Imran Nazar, 2005; released under the BSD public licence.     *
**************************************************************************/

#include <string>
#include <utility>
#include "arm7tdmi.h"
#include "armdasm.h"
#include "plgmmu32.h"
#include "plggui.h"
#include "datadefs.h"
#include "log.h"
#include "err.h"
#include "font5x7.h"

#ifdef ARM7TDMI_X86
# include "x86emit.h"
#endif

//---Static private class members------------------------------------------
// Every plugin has an INFO structure attached, with info about the plugin.

PLUGININFO ARM7TDMI::pInfo={
    PLUGIN_TYPE_CPU,
    0x00010001,
#ifdef ARM7TDMI_X86
    "Dynamic binary translator: ARM7TDMI to x86",
#else
    "Portable ARM7TDMI core",
#endif
    "DSemu-ng"
};

ARM7TDMI::REGS ARM7TDMI::reg;
std::map<u32, int> ARM7TDMI::bkpts;
std::map<u32, CODEBLOCK> ARM7TDMI::blocks;

std::string ARM7TDMI::pluginName;
FILE *ARM7TDMI::dumpfile;
MMU32Plugin *ARM7TDMI::MMU=NULL;
GUIPlugin *ARM7TDMI::GUI=NULL;
u32 *ARM7TDMI::dbgbuffer = NULL;
int ARM7TDMI::dbgwinID = 0;
ARMDasm ARM7TDMI::dasm(rdWhelper);

const char *ARM7TDMI::modeStrings[7] = {
    "USR", "FIQ", "IRQ", "SVC", "ABT", "UND", "SYS"
};

const int ARM7TDMI::modeToCpsrLUT[7] = {
    CPSR_MUSR, CPSR_MFIQ, CPSR_MIRQ, CPSR_MSVC,
    CPSR_MABT, CPSR_MUND, CPSR_MSYS
};

const int ARM7TDMI::modeFromCpsrLUT[16] = {
    MODE_USR, MODE_FIQ, MODE_IRQ, MODE_SVC,
    -1,       -1,       -1,       MODE_ABT,
    -1,       -1,       -1,       MODE_UND,
    -1,       -1,       -1,       MODE_SYS,
};

//---Implementation--------------------------------------------------------
// The ARM7TDMI_DEBUG define masks the raw register dumper; when enabled,
// this will write all 16 registers and the CPSR to a file, in binary,
// every opcode. By default, this is off, because it's godawful slow.

// Initialise plugin
// Parameters: name  - FQPN of plugin as listed in INI file
//             req   - Pointer to PluginRequest API function
//             unreq - Pointer to PluginUnrequest API function
ARM7TDMI::ARM7TDMI(std::string name, REQPTR req, UNREQPTR unreq)
{
    pName = std::string(name);
    pClass = pName.substr(0, pName.find(".")+1);
    pRequest = req;
    pUnrequest = unreq;

    pluginName = pName;

    dbgbuffer = new u32[456*128];

    if(!MMU) MMU = (MMU32Plugin*) pRequest(pClass+"mmu");
    if(!GUI) GUI = (GUIPlugin*) pRequest("UI");

    dbgwinID = GUI->subwinCreate(456, 128, "ARM7 debugger", dbgbuffer);
    
#ifdef ARM7TDMI_DEBUG
    dumpfile = fopen("dsemu-trace.bin","wb");
#endif
    
    MMU->setCPU(this);

    // Simple check: Is this a 32-bit MMU? Our CPU's 32-bit after all
    if((MMU->getCaps() & PLGMMU_CAPS_SIZEMASK) != PLGMMU_CAPS_32B)
    {
        pUnrequest(pClass+"mmu", 1); MMU = NULL;
        throw Exception(ERR_CPU_INIT, pName,
                        pClass+"mmu isn't 32-bit.");
    }

    Logger::log(pName) << "Initialised.";
}

// Plugin cleanup
ARM7TDMI::~ARM7TDMI()
{
    if(dbgbuffer) { delete dbgbuffer; dbgbuffer = NULL; }
    pUnrequest("UI",0); GUI = NULL;

    pUnrequest(pClass+"mmu", 1); MMU = NULL;

#ifdef ARM7TDMI_DEBUG
    fclose(dumpfile);
#endif

    Logger::log(pName) << "Shutdown after " << reg.clkcount << " cycles.";
}

// Reset CPU state
void ARM7TDMI::reset()
{
    memset(&reg, 0, sizeof(REGS));
    reg.curmode = MODE_SYS;
    reg.cpsr = CPSR_MSYS;
	
    MMU->reset();

    memset(dbgbuffer, 0, 456*128*4);
    status(3,0);

#ifdef ARM7TDMI_DEBUG
    reg.r[15]+=4; fwrite(reg.r, 4, 17, dumpfile); reg.r[15]-=4;
#endif

    Logger::log(pName) << "Reset.";
}

// Write to the debugger's framebuffer
void ARM7TDMI::status(int mode=3, int opt2=0)
{
    char str[512]; int a; int xdim=456, ydim=128;
    u32 r15=reg.r[15],r15orig=r15; static u32 r15old;
    static u32 dbgoldr[16],rchanged[16];
    static int dbgoldf[6],fchanged[6];
    u32 op=0; u32 col; u16 oph=0;

    int offset=0;

    if(dbgbuffer)
    {
        Font5x7 prn(dbgbuffer, xdim, ydim);

	cpsrUpdate();

    if((mode==3 && reg.flags[FLAG_T]) || (mode==2))
	r15-=(16-offset*2);
    else
	r15-=(32-offset*4);

    for(a=0;a<16;a++)
    {
        if(r15==r15orig) col=0x00FF8080; else col=0x00FFFFFF;
	if(bkpts.find(r15) != bkpts.end()) col=0x0000FFFF;
        if((r15&0x0F000000)==(r15orig&0x0F000000)) op=MMU->rdW(r15);
        switch(mode)
        {
            case 1:
                if((r15&0x0F000000)==(r15orig&0x0F000000))
                    sprintf(str,"%08X: %08X | %s",r15,op,dasm.disasm(op,r15).c_str());
                else sprintf(str,"%08X",r15);
                r15+=4;
		break;
            case 2:
                if((r15&0x0F000000)==(r15orig&0x0F000000))
                {
                    if(r15&2) oph=op>>16; else oph=op&65535;
                    sprintf(str,"%08X: %04X | %s",r15,oph,dasm.tdisasm((u32)oph,r15).c_str());
                }
                else sprintf(str,"%08X",r15);
                r15+=2; break;
            case 3:
                switch(reg.flags[FLAG_T])
                {
                    case 0:
                        if((r15&0x0F000000)==(r15orig&0x0F000000))
                            sprintf(str,"%08X: %08X | %s",r15,op,dasm.disasm(op,r15).c_str());
                        else sprintf(str,"%08X",r15);
                        r15+=4; break;
                    case 1:
                        if((r15&0x0F000000)==(r15orig&0x0F000000))
                        {
                            if(r15&2) oph=op>>16; else oph=op&65535;
                            sprintf(str,"%08X: %04X | %s",r15,oph,dasm.tdisasm(oph,r15).c_str());
                        }
                        else sprintf(str,"%08X",r15);
                        r15+=2; break;
                }
        }
        prn.print(str,0,a*8,col);
    }

    for(a=0;a<=15;a++)
    {
        sprintf(str,"r%02d: %08X",a,reg.r[a]);
        if(r15old!=r15orig)
        {
            if(reg.r[a]!=dbgoldr[a]) { dbgoldr[a]=reg.r[a]; rchanged[a]=1; }
            else rchanged[a]=0;
        }
        if(rchanged[a]) prn.print(str,56*6,a*8,0x000000FF);
	else prn.print(str,56*6,a*8,0x00FFFFFF);
    }
    
    sprintf(str,"%s",modeStrings[reg.curmode]);
    prn.print(str, 72*6, 56 ,0x00FFFFFF);

    a=(reg.cpsr&CPSR_N)>>31; sprintf(str,"N: %d",a);
    if(r15old!=r15orig)
        if(a!=dbgoldf[0]) { dbgoldf[0]=a; fchanged[0]=1; } else fchanged[0]=0;
    if(fchanged[0]) prn.print(str,72*6,0,0x000000FF);
    else prn.print(str,72*6,0,0x00FFFFFF);

    a=(reg.cpsr&CPSR_Z)>>30; sprintf(str,"Z: %d",a);
    if(r15old!=r15orig)
        if(a!=dbgoldf[1]) { dbgoldf[1]=a; fchanged[1]=1; } else fchanged[1]=0;
    if(fchanged[1]) prn.print(str,72*6,8,0x000000FF);
    else prn.print(str,72*6,8,0x00FFFFFF);

    a=(reg.cpsr&CPSR_C)>>29; sprintf(str,"C: %d",a);
    if(r15old!=r15orig)
        if(a!=dbgoldf[2]) { dbgoldf[2]=a; fchanged[2]=1; } else fchanged[2]=0;
    if(fchanged[2]) prn.print(str,72*6,16,0x000000FF);
    else prn.print(str,72*6,16,0x00FFFFFF);

    a=(reg.cpsr&CPSR_V)>>28; sprintf(str,"V: %d",a);
    if(r15old!=r15orig)
        if(a!=dbgoldf[3]) { dbgoldf[3]=a; fchanged[3]=1; } else fchanged[3]=0;
    if(fchanged[3]) prn.print(str,72*6,24,0x000000FF);
    else prn.print(str,72*6,24,0x00FFFFFF);

    a=(reg.cpsr&CPSR_T)>>5;  sprintf(str,"T: %d",a);
    if(r15old!=r15orig)
        if(a!=dbgoldf[5]) { dbgoldf[5]=a; fchanged[5]=1; } else fchanged[5]=0;
    if(fchanged[5]) prn.print(str,72*6,40,0x000000FF);
    else prn.print(str,72*6,40,0x00FFFFFF);

    if(r15old!=r15orig) r15old=r15orig;

    }

    GUI->subwinRefresh(dbgwinID);
}

// Function the disassembler will use to read memory
u32 ARM7TDMI::rdWhelper(u32 addr) { return MMU->rdW(addr); }

// Save registers, and switch processor modes
void ARM7TDMI::modeSwitch(int in, int out)
{
    //---Save state
    reg.r0 = reg.r[0]; reg.r1 = reg.r[1];
    reg.r2 = reg.r[2]; reg.r3 = reg.r[3];
    reg.r4 = reg.r[4]; reg.r5 = reg.r[5];
    reg.r6 = reg.r[6]; reg.r7 = reg.r[7];

    if(in == MODE_FIQ)
    {
	reg.r8fiq = reg.r[8];   reg.r9fiq = reg.r[9];
	reg.r10fiq = reg.r[10]; reg.r11fiq = reg.r[11];
	reg.r12fiq = reg.r[12];
    } else {
	reg.r8 = reg.r[8];   reg.r9 = reg.r[9];
	reg.r10 = reg.r[10]; reg.r11 = reg.r[11];
	reg.r12 = reg.r[12];
    }
    switch(in)
    {
	case MODE_USR:
	case MODE_SYS: reg.r13 = reg.r[13]; reg.r14 = reg.r[14]; break;
	case MODE_IRQ: reg.r13irq = reg.r[13]; reg.r14irq = reg.r[14]; break;
	case MODE_FIQ: reg.r13fiq = reg.r[13]; reg.r14fiq = reg.r[14]; break;
	case MODE_SVC: reg.r13svc = reg.r[13]; reg.r14svc = reg.r[14]; break;
	case MODE_ABT: reg.r13abt = reg.r[13]; reg.r14abt = reg.r[14]; break;
	case MODE_UND: reg.r13und = reg.r[13]; reg.r14und = reg.r[14]; break;
    }
    reg.r15 = reg.r[15];
    
    //---Switch to new mode
    if(out == MODE_FIQ)
    {
	reg.r[8] = reg.r8fiq;   reg.r[9] = reg.r9fiq;
	reg.r[10] = reg.r10fiq; reg.r[11] = reg.r11fiq;
	reg.r[12] = reg.r12fiq;
    } else {
	reg.r[8] = reg.r8;   reg.r[9] = reg.r9;
	reg.r[10] = reg.r10; reg.r[11] = reg.r11;
	reg.r[12] = reg.r12;
    }
    switch(out)
    {
	case MODE_USR:
	case MODE_SYS: reg.r[13] = reg.r13; reg.r[14] = reg.r14; break;
	case MODE_IRQ: reg.r[13] = reg.r13irq; reg.r[14] = reg.r14irq; break;
	case MODE_FIQ: reg.r[13] = reg.r13fiq; reg.r[14] = reg.r14fiq; break;
	case MODE_SVC: reg.r[13] = reg.r13svc; reg.r[14] = reg.r14svc; break;
	case MODE_ABT: reg.r[13] = reg.r13abt; reg.r[14] = reg.r14abt; break;
	case MODE_UND: reg.r[13] = reg.r13und; reg.r[14] = reg.r14und; break;
    }
    
    reg.curmode = out;
    cpsrUpdate();
}

// Update the CPSR based on the flags[] array
void ARM7TDMI::cpsrUpdate()
{
    if(reg.flags[FLAG_N]) reg.cpsr|=CPSR_N; else reg.cpsr&=(~CPSR_N);
    if(reg.flags[FLAG_Z]) reg.cpsr|=CPSR_Z; else reg.cpsr&=(~CPSR_Z);
    if(reg.flags[FLAG_C]) reg.cpsr|=CPSR_C; else reg.cpsr&=(~CPSR_C);
    if(reg.flags[FLAG_V]) reg.cpsr|=CPSR_V; else reg.cpsr&=(~CPSR_V);
    if(reg.flags[FLAG_T]) reg.cpsr|=CPSR_T; else reg.cpsr&=(~CPSR_T);
    reg.cpsr &= ~CPSR_M;
    reg.cpsr |= modeToCpsrLUT[reg.curmode];
}

// Split the CPSR into the flags[] array
void ARM7TDMI::cpsrSplit()
{
    reg.flags[FLAG_N]=(reg.cpsr>>31)&1;
    reg.flags[FLAG_Z]=(reg.cpsr>>30)&1;
    reg.flags[FLAG_C]=(reg.cpsr>>29)&1;
    reg.flags[FLAG_V]=(reg.cpsr>>28)&1;
    reg.flags[FLAG_T]=(reg.cpsr>>5)&1;

    modeSwitch(reg.curmode, modeFromCpsrLUT[reg.cpsr&15]);
}

// Add to the internal clock
void ARM7TDMI::clockAdd(int cyc)
{
    reg.clkcount += cyc;
}

// Toggle a breakpoint on or off at a specific address
void ARM7TDMI::dbgBkptToggle(u32 addr)
{
    if(bkpts.find(addr) != bkpts.end()) bkpts.erase(addr);
    else bkpts[addr]=1;

    status();
}

// Return the map of breakpoint addresses to states
std::map<u32, int> ARM7TDMI::dbgBkptGet()
{
    return bkpts;
}

// Work out how many bits in a 16-bit value are set to 1
// Used by the LDM/STM opcodes
// ACK: Donald Knuth (Sidesum algorithm)
int ARM7TDMI::ssum16(u16 val)
{
    int res=val;
    res=((res>>1)&0x5555)+(res&0x5555);
    res=((res>>2)&0x3333)+(res&0x3333);
    res=((res>>4)&0x0F0F)+(res&0x0F0F);
    res=((res>>8)&0x00FF)+(res&0x00FF);
    return res;
}

// CPU is interrupted; set up the CPU state for the int handler
void ARM7TDMI::interrupt(int type)
{
    // Type ignored for now, IRQ assumed
    reg.flags[FLAG_T]=0;
    cpsrUpdate();
    reg.cpsr |= CPSR_I;
    
    reg.r14irq = reg.r[15];
    reg.spsr[MODE_IRQ] = reg.cpsr;
    modeSwitch(reg.curmode, MODE_IRQ);

    reg.r[15] = 0x00000018;
    reg.blockstart = 0x00000018;
}

// Set the processor's PC from outside
void ARM7TDMI::setPC(u32 pc)
{
    reg.r[15] = pc;
    reg.blockstart = pc;

    // Neutralise the translation, to accommodate SMC
    if(blocks.find(reg.r[15]) != blocks.end())
	blocks[reg.r[15]].target_addr = NULL;
}

// Internal function: execute one instruction
inline int ARM7TDMI::execOne(uint64_t start)
{
	// Check the bkpts map to see if we hit one
        if(bkpts.find(reg.r[15]) != bkpts.end())
        {
            switch(bkpts[reg.r[15]])
            {
                case 1:
                    bkpts[reg.r[15]]=2;
                    return -(reg.clkcount-start);
                case 2:
                    bkpts[reg.r[15]]=1;
                    break;
            }
        }
	
	if(reg.flags[FLAG_T])
	{
	    reg.curop = (u32)MMU->rdH(reg.r[15]);
	    reg.r[15] += 2;

	    reg.clkcount += TLUT[(reg.curop&TMSK_OP)>>TSHFT_OP]();

#ifdef ARM7TDMI_DEBUG
	    cpsrUpdate();
            reg.r[15]+=2; fwrite(reg.r, 4, 17, dumpfile); reg.r[15]-=2;
#endif	    
	    
	} else {
            reg.curop = MMU->rdW(reg.r[15]);
	    
	    reg.r[15] += 4;
	    
            if(condLUT[(reg.curop&MSK_COND)>>SHFT_COND]())
            {
                reg.clkcount += OLUT[((reg.curop&MSK_OP)>>SHFT_OPH)|
                                     ((reg.curop&MSK_FUNC)>>SHFT_FUNC)]();
            }
            else reg.clkcount++;
	    
#ifdef ARM7TDMI_DEBUG
	    cpsrUpdate();
            reg.r[15]+=4; fwrite(reg.r, 4, 17, dumpfile); reg.r[15]-=4;
#endif	    
	    
	}
	return reg.clkcount-start;
}

//---Dispatch--------------------------------------------------------------
// TODO: Optimise the dispatch loop, it's damned slow

// If you want the dynarec, define this.
#ifdef ARM7TDMI_X86

void ARM7TDMI::translate(CODEBLOCK *block)
{
    int blockdone = 0, clk = 0;
    u32 curop, r15;
	
    block->target_addr = (u8*)malloc(4096);
    if(!block->target_addr)
	throw Exception(ERR_CPU_DYNALLOC, pluginName,
			"Couldn't allocate space for a new codeblock");

//    printf("Target block allocated at %08X\n", block->target_addr);
    
    x86::setblock(block);
    r15 = block->ARM_startaddr;
    
    do {
	if(block->flags & BLOCK_THUMB)
	{
	    curop = (u32)MMU->rdH(r15);
	    clk += XTLUT[(curop&TMSK_OP)>>TSHFT_OP]();
	    
	} else {
	
            curop = MMU->rdW(r15);
	    // TODO: Conditional execution, how in the hell.
	    
//            printf("Decoding %08X with function %08X\n", curop, XOLUT[((curop&MSK_OP)>>SHFT_OPH)|((curop&MSK_FUNC)>>SHFT_FUNC)]);
                clk += XOLUT[((curop&MSK_OP)>>SHFT_OPH)|((curop&MSK_FUNC)>>SHFT_FUNC)]();
	}
	
	if(block->flags & BLOCK_THUMB) r15+=2; else r15+=4;
        if(r15 > block->ARM_endaddr) blockdone = 1;
    } while(!blockdone);
    
    x86::RET(clk);
}

int ARM7TDMI::exec(int cycles)
{
    uint64_t stamp = reg.clkcount+cycles, start = reg.clkcount;

    // Handle the special case for single-stepping
    if(cycles == 1) 
	return execOne(start);

    do {
/*	if(blocks.find(reg.r[15]) != blocks.end())
	{
	    // The interpreter's found a block, and filled in a
	    // CODEBLOCK structure. If it's not translated yet,
	    // do that, and then use it.

//	    printf("Block found for r15=%08X to %08X, x86=%08X\n", reg.r[15], blocks[reg.r[15]].ARM_endaddr, blocks[reg.r[15]].target_addr);
		
            // If a breakpoint is found within this block, drop out now.
	    for(std::map<u32,int>::iterator i=bkpts.begin(); i!=bkpts.end(); ++i)
	    {
		if(((*i).first >= reg.r[15]) && 
		   ((*i).first <= blocks[reg.r[15]].ARM_endaddr))
		{
		    return -1;
		}
	    }
	    
	    if(!blocks[reg.r[15]].target_addr)
		translate(&blocks[reg.r[15]]);

	    blocks[reg.r[15]].timestamp = reg.clkcount;
	    reg.clkcount += ((opfptr)(blocks[reg.r[15]].target_addr))();

//	    printf("That block went fine.\n");
	    
	}
	else
	{
*/	    uint64_t blockstart = reg.clkcount;	
	    int blockdone=0;
	    
	    do {
	        u32 r15=reg.r[15], tmode=reg.flags[FLAG_T];
		
		// Run an instruction. If it changed the PC, that's the end
		// of this block, so save info to a new CODEBLOCK.
		execOne(start);
		if((reg.r[15] != (r15+4)) || (tmode != reg.flags[FLAG_T]))
		{
		//    char str[128];
		//    sprintf(str,"Block end detected. %08X to %08X for %d cycles", reg.blockstart, r15, reg.clkcount-blockstart);
		//    Logger::log(pluginName) << str;

		    CODEBLOCK blk={0};
		    blk.target_addr=NULL;
		    blk.ARM_startaddr=reg.blockstart;
		    blk.ARM_endaddr=r15;
		    blk.ARM_cycles=reg.clkcount-blockstart;
		    if(tmode) blk.flags |= BLOCK_THUMB; else blk.flags &= ~BLOCK_THUMB;

		    blocks[reg.blockstart] = blk;
		    reg.blockstart = reg.r[15];
		    blockdone = 1;
		}
	    } while(!blockdone);
//	}
    } while(reg.clkcount < stamp);
    return reg.clkcount - start;
}

// If you want the plain old interpreter, don't define it.
#else

int ARM7TDMI::exec(int cycles)
{
    uint64_t stamp = reg.clkcount+cycles, start = reg.clkcount;
    do execOne(start); while(reg.clkcount < stamp);
    return reg.clkcount - start;
}

#endif//ARM7TDMI_X86

//---Exceptional opcode handlers-------------------------------------------
// NOTE: As of right now, these do nothing; they are there such that the
//       opcodes can safely call them if a problem occurs. 

OPC opUND()
{
    //char str[80];
    //sprintf(str, "%08X: %08X", reg.r[15], reg.curop);
    //Logger::log(pluginName) << "Undefined opcode at " << str;
    return 1;
}

OPC opUNP()
{
    //char str[80];
    //sprintf(str, "%08X: %08X", reg.r[15], reg.curop);
    //Logger::log(pluginName) << "Unpredictable opcode at " << str;
    return 1;
}

OPC opUNI()
{
    //char str[80];
    //sprintf(str, "%08X: %08X", reg.r[15], reg.curop);
    //Logger::log(pluginName) << "Unimplemented opcode at " << str;
    return 1;
}

OPC opUNL()
{
    // Would act as cache filler
    return 1;
}

#ifdef ARM7TDMI_X86

OPC xopUND() { return opUND(); }
OPC xopUNP() { return opUNP(); }
OPC xopUNI() { return opUNI(); }
OPC xopUNL() { return opUNL(); }

#endif

//---Plugin Support--------------------------------------------------------
// Retrieve Plugin class from outside
// Parameters: plg   - Address of a pointer to a Plugin class to 'new'
//             name  - FQPN of plugin as listed in INI file
//             req   - Pointer to PluginRequest API function
//             unreq - Pointer to PluginUnrequest API function
EXPORTFUNC void getPlugin(Plugin **plg,
                          std::string name,
                          REQPTR req, UNREQPTR unreq)
{
    // Initialise a new Test plugin at the parameter
    *plg = new ARM7TDMI(name, req, unreq);
}

// Provide plugin version information
PLUGININFO *ARM7TDMI::getinfo()
{
    return &pInfo;
}

// Release plugin from outside
void ARM7TDMI::release()
{
    // Delete the Test plugin that was 'new'd in getPlugin.
    delete this;
}

/*** EOF: arm7tdmi.cpp ***************************************************/

