/**************************************************************************
* DSemu - The Next Generation                                             *
* Portable ARM9ES core: Class definition [arm9es.h]                       *
* Copyright Imran Nazar, 2005; released under the BSD public licence.     *
**************************************************************************/

#ifndef __ARM9ES_H_
#define __ARM9ES_H_

#include <utility>
#include <map>
#include <stdio.h>
#include "defs.h"
#include "plgcpu.h"
#include "plgmmu32.h"
#include "plggui.h"
#include "datadefs.h"
#include "armdasm.h"
#include "armcopro.h"

// An implementation of the CPU plugin interface.
class ARM9ES : public CPUPlugin
{
    public:
        ARM9ES(std::string name,REQPTR req, UNREQPTR unreq);
        ~ARM9ES();

        PLUGININFO *getinfo();
        void reset();
        void release();
        void status(int,int);

        int exec(int);
	void clockAdd(int);
	void interrupt(int);
	
	void dbgBkptToggle(u32);
	std::map<u32, int> dbgBkptGet();

	void setPC(u32);

    private:

        //---Plugin-generic stuff
        std::string pName, pClass;
        REQPTR pRequest;
        UNREQPTR pUnrequest;
	static PLUGININFO pInfo;
        
        static std::string pluginName;

	//---Start of the stuff we'll be needing
        static MMU32Plugin *MMU;
	static GUIPlugin *GUI;
	
	static u32 *dbgbuffer;
	static int dbgwinID;
        static ARMDasm dasm;
	static u32 rdWhelper(u32);

        static FILE *dumpfile;
	
        //---Register set
        typedef struct {
            u32 r[16], cpsr, spsr[7], flags[8];
            u32 curop, curmode;
            uint64_t clkcount;

            u32 r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11,r12,r13,r14,r15;
            u32 r8fiq, r9fiq, r10fiq, r11fiq, r12fiq, r13fiq, r14fiq;
            u32 r13svc, r14svc, r13abt, r14abt, r13irq, r14irq, r13und, r14und;

            u32 tmp1, tmp2, tmp3, tmp4, tmpc;
	    uint64_t tmp64_1, tmp64_2;
        } REGS;

        static REGS reg;

        static std::map<u32, int> bkpts;

	//---Internal flag to/from CPSR helpers
        static void cpsrSplit(), cpsrUpdate();
	
        //---Coprocessors
	static ARMCopro *copro[16];
	
        //---Opcode tables
        typedef int OPCODE;
        typedef OPCODE (*opfptr)();
	typedef int (*intfptr)();

        static const opfptr OLUT[4096];
	static const opfptr TLUT[256];
	static const intfptr condLUT[16];

	//---Mode-switching
	static const char *modeStrings[7];
	static const int modeToCpsrLUT[7];
	static const int modeFromCpsrLUT[16];
	static void modeSwitch(int, int);

        //---Conditional execution helpers
	static int condEQ(), condNE(), condCS(), condCC();
	static int condMI(), condPL(), condVS(), condVC();
	static int condHI(), condLS(), condGE(), condLT();
	static int condGT(), condLE(), condAL(), condNV();
	
        //---Instructions

	// Helper functions
	static u32 nlpo2(u32);
	static int ssum16(u16);
        
	static OPCODE opUNP();   // Unpredictable op
        static OPCODE opUND();   // Undefined op
        static OPCODE opUNI();   // Unimplemented op
	static OPCODE opUNL();   // Unlinked (opcode cache filler)
	
        //---Branch (with Link)----------------------------------------------------

        static OPCODE opB(), opBL();
        static OPCODE opBX();
	
        static OPCODE opBLX(), opBLXr();

        //---DP (DataProcessing); LsL/LsR/AsR/RoR; Immediate/Register--------------

        static OPCODE opANDlli(), opANDllr();
        static OPCODE opANDlri(), opANDlrr();
        static OPCODE opANDari(), opANDarr();
        static OPCODE opANDrri(), opANDrrr();
        static OPCODE opANDimm();

        static OPCODE opEORlli(), opEORllr();
        static OPCODE opEORlri(), opEORlrr();
        static OPCODE opEORari(), opEORarr();
        static OPCODE opEORrri(), opEORrrr();
        static OPCODE opEORimm();

        static OPCODE opSUBlli(), opSUBllr();
        static OPCODE opSUBlri(), opSUBlrr();
        static OPCODE opSUBari(), opSUBarr();
        static OPCODE opSUBrri(), opSUBrrr();
        static OPCODE opSUBimm();

        static OPCODE opRSBlli(), opRSBllr();
        static OPCODE opRSBlri(), opRSBlrr();
        static OPCODE opRSBari(), opRSBarr();
        static OPCODE opRSBrri(), opRSBrrr();
        static OPCODE opRSBimm();

        static OPCODE opADDlli(), opADDllr();
        static OPCODE opADDlri(), opADDlrr();
        static OPCODE opADDari(), opADDarr();
        static OPCODE opADDrri(), opADDrrr();
        static OPCODE opADDimm();

        static OPCODE opADClli(), opADCllr();
        static OPCODE opADClri(), opADClrr();
        static OPCODE opADCari(), opADCarr();
        static OPCODE opADCrri(), opADCrrr();
        static OPCODE opADCimm();

        static OPCODE opSBClli(), opSBCllr();
        static OPCODE opSBClri(), opSBClrr();
        static OPCODE opSBCari(), opSBCarr();
        static OPCODE opSBCrri(), opSBCrrr();
        static OPCODE opSBCimm();

        static OPCODE opRSClli(), opRSCllr();
        static OPCODE opRSClri(), opRSClrr();
        static OPCODE opRSCari(), opRSCarr();
        static OPCODE opRSCrri(), opRSCrrr();
        static OPCODE opRSCimm();

        static OPCODE opORRlli(), opORRllr();
        static OPCODE opORRlri(), opORRlrr();
        static OPCODE opORRari(), opORRarr();
        static OPCODE opORRrri(), opORRrrr();
        static OPCODE opORRimm();

        static OPCODE opMOVlli(), opMOVllr();
        static OPCODE opMOVlri(), opMOVlrr();
        static OPCODE opMOVari(), opMOVarr();
        static OPCODE opMOVrri(), opMOVrrr();
        static OPCODE opMOVimm();

        static OPCODE opBIClli(), opBICllr();
        static OPCODE opBIClri(), opBIClrr();
        static OPCODE opBICari(), opBICarr();
        static OPCODE opBICrri(), opBICrrr();
        static OPCODE opBICimm();

        static OPCODE opMVNlli(), opMVNllr();
        static OPCODE opMVNlri(), opMVNlrr();
        static OPCODE opMVNari(), opMVNarr();
        static OPCODE opMVNrri(), opMVNrrr();
        static OPCODE opMVNimm();

        static OPCODE opANDSlli(), opANDSllr();
        static OPCODE opANDSlri(), opANDSlrr();
        static OPCODE opANDSari(), opANDSarr();
        static OPCODE opANDSrri(), opANDSrrr();
        static OPCODE opANDSimm();

        static OPCODE opEORSlli(), opEORSllr();
        static OPCODE opEORSlri(), opEORSlrr();
        static OPCODE opEORSari(), opEORSarr();
        static OPCODE opEORSrri(), opEORSrrr();
        static OPCODE opEORSimm();

        static OPCODE opSUBSlli(), opSUBSllr();
        static OPCODE opSUBSlri(), opSUBSlrr();
        static OPCODE opSUBSari(), opSUBSarr();
        static OPCODE opSUBSrri(), opSUBSrrr();
        static OPCODE opSUBSimm();

        static OPCODE opRSBSlli(), opRSBSllr();
        static OPCODE opRSBSlri(), opRSBSlrr();
        static OPCODE opRSBSari(), opRSBSarr();
        static OPCODE opRSBSrri(), opRSBSrrr();
        static OPCODE opRSBSimm();

        static OPCODE opADDSlli(), opADDSllr();
        static OPCODE opADDSlri(), opADDSlrr();
        static OPCODE opADDSari(), opADDSarr();
        static OPCODE opADDSrri(), opADDSrrr();
        static OPCODE opADDSimm();

        static OPCODE opADCSlli(), opADCSllr();
        static OPCODE opADCSlri(), opADCSlrr();
        static OPCODE opADCSari(), opADCSarr();
        static OPCODE opADCSrri(), opADCSrrr();
        static OPCODE opADCSimm();

        static OPCODE opSBCSlli(), opSBCSllr();
        static OPCODE opSBCSlri(), opSBCSlrr();
        static OPCODE opSBCSari(), opSBCSarr();
        static OPCODE opSBCSrri(), opSBCSrrr();
        static OPCODE opSBCSimm();

        static OPCODE opRSCSlli(), opRSCSllr();
        static OPCODE opRSCSlri(), opRSCSlrr();
        static OPCODE opRSCSari(), opRSCSarr();
        static OPCODE opRSCSrri(), opRSCSrrr();
        static OPCODE opRSCSimm();

        static OPCODE opTSTSlli(), opTSTSllr();
        static OPCODE opTSTSlri(), opTSTSlrr();
        static OPCODE opTSTSari(), opTSTSarr();
        static OPCODE opTSTSrri(), opTSTSrrr();
        static OPCODE opTSTSimm();

        static OPCODE opTEQSlli(), opTEQSllr();
        static OPCODE opTEQSlri(), opTEQSlrr();
        static OPCODE opTEQSari(), opTEQSarr();
        static OPCODE opTEQSrri(), opTEQSrrr();
        static OPCODE opTEQSimm();

        static OPCODE opCMPSlli(), opCMPSllr();
        static OPCODE opCMPSlri(), opCMPSlrr();
        static OPCODE opCMPSari(), opCMPSarr();
        static OPCODE opCMPSrri(), opCMPSrrr();
        static OPCODE opCMPSimm();

        static OPCODE opCMNSlli(), opCMNSllr();
        static OPCODE opCMNSlri(), opCMNSlrr();
        static OPCODE opCMNSari(), opCMNSarr();
        static OPCODE opCMNSrri(), opCMNSrrr();
        static OPCODE opCMNSimm();

        static OPCODE opORRSlli(), opORRSllr();
        static OPCODE opORRSlri(), opORRSlrr();
        static OPCODE opORRSari(), opORRSarr();
        static OPCODE opORRSrri(), opORRSrrr();
        static OPCODE opORRSimm();

        static OPCODE opMOVSlli(), opMOVSllr();
        static OPCODE opMOVSlri(), opMOVSlrr();
        static OPCODE opMOVSari(), opMOVSarr();
        static OPCODE opMOVSrri(), opMOVSrrr();
        static OPCODE opMOVSimm();

        static OPCODE opBICSlli(), opBICSllr();
        static OPCODE opBICSlri(), opBICSlrr();
        static OPCODE opBICSari(), opBICSarr();
        static OPCODE opBICSrri(), opBICSrrr();
        static OPCODE opBICSimm();

        static OPCODE opMVNSlli(), opMVNSllr();
        static OPCODE opMVNSlri(), opMVNSlrr();
        static OPCODE opMVNSari(), opMVNSarr();
        static OPCODE opMVNSrri(), opMVNSrrr();
        static OPCODE opMVNSimm();

        //---Multiply/LongMul------------------------------------------------------

        static OPCODE opCLZ();
	
        static OPCODE opMUL(), opMLA();
        static OPCODE opUMULL(), opUMLAL(), opSMULL(), opSMLAL();

        static OPCODE opMULS(), opMLAS();
        static OPCODE opUMULLS(), opUMLALS(), opSMULLS(), opSMLALS();

        //---Move Register From Status; Imm/Reg; Cpsr/Spsr-------------------------

        static OPCODE opMRSrc(), opMRSrs();

        //---Move Status from Register; Cpsr/Spsr----------------------------------

        static OPCODE opMSRrc(), opMSRrs();
        static OPCODE opMSRic(), opMSRis();

        //---LS (Load/Store); PReidx/OFfset/PsTidx; Imm/Reg; Minus/Plus; (Shift)---

        static OPCODE opLDRptim(), opLDRofim(), opLDRprim();
        static OPCODE opLDRptip(), opLDRofip(), opLDRprip();

        static OPCODE opSTRptim(), opSTRofim(), opSTRprim();
        static OPCODE opSTRptip(), opSTRofip(), opSTRprip();

        static OPCODE opLDRBptim(), opLDRBofim(), opLDRBprim();
        static OPCODE opLDRBptip(), opLDRBofip(), opLDRBprip();

        static OPCODE opSTRBptim(), opSTRBofim(), opSTRBprim();
        static OPCODE opSTRBptip(), opSTRBofip(), opSTRBprip();

        static OPCODE opLDRHptim(), opLDRHofim(), opLDRHprim();
        static OPCODE opLDRHptip(), opLDRHofip(), opLDRHprip();
        static OPCODE opLDRHptrm(), opLDRHofrm(), opLDRHprrm();
        static OPCODE opLDRHptrp(), opLDRHofrp(), opLDRHprrp();

        static OPCODE opSTRHptim(), opSTRHofim(), opSTRHprim();
        static OPCODE opSTRHptip(), opSTRHofip(), opSTRHprip();
        static OPCODE opSTRHptrm(), opSTRHofrm(), opSTRHprrm();
        static OPCODE opSTRHptrp(), opSTRHofrp(), opSTRHprrp();

        static OPCODE opLDRTptim(), opLDRTofim(), opLDRTprim();
        static OPCODE opLDRTptip(), opLDRTofip(), opLDRTprip();

        static OPCODE opSTRTptim(), opSTRTofim(), opSTRTprim();
        static OPCODE opSTRTptip(), opSTRTofip(), opSTRTprip();

        static OPCODE opLDRBTptim(), opLDRBTofim(), opLDRBTprim();
        static OPCODE opLDRBTptip(), opLDRBTofip(), opLDRBTprip();

        static OPCODE opSTRBTptim(), opSTRBTofim(), opSTRBTprim();
        static OPCODE opSTRBTptip(), opSTRBTofip(), opSTRBTprip();

        static OPCODE opLDRSHptim(), opLDRSHofim(), opLDRSHprim();
        static OPCODE opLDRSHptip(), opLDRSHofip(), opLDRSHprip();
        static OPCODE opLDRSHptrm(), opLDRSHofrm(), opLDRSHprrm();
        static OPCODE opLDRSHptrp(), opLDRSHofrp(), opLDRSHprrp();

        static OPCODE opLDRSBptim(), opLDRSBofim(), opLDRSBprim();
        static OPCODE opLDRSBptip(), opLDRSBofip(), opLDRSBprip();
        static OPCODE opLDRSBptrm(), opLDRSBofrm(), opLDRSBprrm();
        static OPCODE opLDRSBptrp(), opLDRSBofrp(), opLDRSBprrp();

        static OPCODE opLDRptrmll(), opLDRofrmll(), opLDRprrmll();
        static OPCODE opLDRptrpll(), opLDRofrpll(), opLDRprrpll();
        static OPCODE opLDRptrmlr(), opLDRofrmlr(), opLDRprrmlr();
        static OPCODE opLDRptrplr(), opLDRofrplr(), opLDRprrplr();
        static OPCODE opLDRptrmar(), opLDRofrmar(), opLDRprrmar();
        static OPCODE opLDRptrpar(), opLDRofrpar(), opLDRprrpar();
        static OPCODE opLDRptrmrr(), opLDRofrmrr(), opLDRprrmrr();
        static OPCODE opLDRptrprr(), opLDRofrprr(), opLDRprrprr();

        static OPCODE opSTRptrmll(), opSTRofrmll(), opSTRprrmll();
        static OPCODE opSTRptrpll(), opSTRofrpll(), opSTRprrpll();
        static OPCODE opSTRptrmlr(), opSTRofrmlr(), opSTRprrmlr();
        static OPCODE opSTRptrplr(), opSTRofrplr(), opSTRprrplr();
        static OPCODE opSTRptrmar(), opSTRofrmar(), opSTRprrmar();
        static OPCODE opSTRptrpar(), opSTRofrpar(), opSTRprrpar();
        static OPCODE opSTRptrmrr(), opSTRofrmrr(), opSTRprrmrr();
        static OPCODE opSTRptrprr(), opSTRofrprr(), opSTRprrprr();

        static OPCODE opLDRBptrmll(), opLDRBofrmll(), opLDRBprrmll();
        static OPCODE opLDRBptrpll(), opLDRBofrpll(), opLDRBprrpll();
        static OPCODE opLDRBptrmlr(), opLDRBofrmlr(), opLDRBprrmlr();
        static OPCODE opLDRBptrplr(), opLDRBofrplr(), opLDRBprrplr();
        static OPCODE opLDRBptrmar(), opLDRBofrmar(), opLDRBprrmar();
        static OPCODE opLDRBptrpar(), opLDRBofrpar(), opLDRBprrpar();
        static OPCODE opLDRBptrmrr(), opLDRBofrmrr(), opLDRBprrmrr();
        static OPCODE opLDRBptrprr(), opLDRBofrprr(), opLDRBprrprr();

        static OPCODE opSTRBptrmll(), opSTRBofrmll(), opSTRBprrmll();
        static OPCODE opSTRBptrpll(), opSTRBofrpll(), opSTRBprrpll();
        static OPCODE opSTRBptrmlr(), opSTRBofrmlr(), opSTRBprrmlr();
        static OPCODE opSTRBptrplr(), opSTRBofrplr(), opSTRBprrplr();
        static OPCODE opSTRBptrmar(), opSTRBofrmar(), opSTRBprrmar();
        static OPCODE opSTRBptrpar(), opSTRBofrpar(), opSTRBprrpar();
        static OPCODE opSTRBptrmrr(), opSTRBofrmrr(), opSTRBprrmrr();
        static OPCODE opSTRBptrprr(), opSTRBofrprr(), opSTRBprrprr();

        static OPCODE opLDRTptrmll(), opLDRTofrmll(), opLDRTprrmll();
        static OPCODE opLDRTptrpll(), opLDRTofrpll(), opLDRTprrpll();
        static OPCODE opLDRTptrmlr(), opLDRTofrmlr(), opLDRTprrmlr();
        static OPCODE opLDRTptrplr(), opLDRTofrplr(), opLDRTprrplr();
        static OPCODE opLDRTptrmar(), opLDRTofrmar(), opLDRTprrmar();
        static OPCODE opLDRTptrpar(), opLDRTofrpar(), opLDRTprrpar();
        static OPCODE opLDRTptrmrr(), opLDRTofrmrr(), opLDRTprrmrr();
        static OPCODE opLDRTptrprr(), opLDRTofrprr(), opLDRTprrprr();

        static OPCODE opSTRTptrmll(), opSTRTofrmll(), opSTRTprrmll();
        static OPCODE opSTRTptrpll(), opSTRTofrpll(), opSTRTprrpll();
        static OPCODE opSTRTptrmlr(), opSTRTofrmlr(), opSTRTprrmlr();
        static OPCODE opSTRTptrplr(), opSTRTofrplr(), opSTRTprrplr();
        static OPCODE opSTRTptrmar(), opSTRTofrmar(), opSTRTprrmar();
        static OPCODE opSTRTptrpar(), opSTRTofrpar(), opSTRTprrpar();
        static OPCODE opSTRTptrmrr(), opSTRTofrmrr(), opSTRTprrmrr();
        static OPCODE opSTRTptrprr(), opSTRTofrprr(), opSTRTprrprr();

        static OPCODE opLDRBTptrmll(), opLDRBTofrmll(), opLDRBTprrmll();
        static OPCODE opLDRBTptrpll(), opLDRBTofrpll(), opLDRBTprrpll();
        static OPCODE opLDRBTptrmlr(), opLDRBTofrmlr(), opLDRBTprrmlr();
        static OPCODE opLDRBTptrplr(), opLDRBTofrplr(), opLDRBTprrplr();
        static OPCODE opLDRBTptrmar(), opLDRBTofrmar(), opLDRBTprrmar();
        static OPCODE opLDRBTptrpar(), opLDRBTofrpar(), opLDRBTprrpar();
        static OPCODE opLDRBTptrmrr(), opLDRBTofrmrr(), opLDRBTprrmrr();
        static OPCODE opLDRBTptrprr(), opLDRBTofrprr(), opLDRBTprrprr();

        static OPCODE opSTRBTptrmll(), opSTRBTofrmll(), opSTRBTprrmll();
        static OPCODE opSTRBTptrpll(), opSTRBTofrpll(), opSTRBTprrpll();
        static OPCODE opSTRBTptrmlr(), opSTRBTofrmlr(), opSTRBTprrmlr();
        static OPCODE opSTRBTptrplr(), opSTRBTofrplr(), opSTRBTprrplr();
        static OPCODE opSTRBTptrmar(), opSTRBTofrmar(), opSTRBTprrmar();
        static OPCODE opSTRBTptrpar(), opSTRBTofrpar(), opSTRBTprrpar();
        static OPCODE opSTRBTptrmrr(), opSTRBTofrmrr(), opSTRBTprrmrr();
        static OPCODE opSTRBTptrprr(), opSTRBTofrprr(), opSTRBTprrprr();

        //---SWP-------------------------------------------------------------------

        static OPCODE opSWP(), opSWPB();

        //---LoaD/STore Multiple; Increment/Decrement After/Before-----------------

        static OPCODE opLDMIA(), opLDMIB(), opLDMDA(), opLDMDB();
        static OPCODE opSTMIA(), opSTMIB(), opSTMDA(), opSTMDB();

        static OPCODE opLDMIAu(), opLDMIBu(), opLDMDAu(), opLDMDBu();
        static OPCODE opSTMIAu(), opSTMIBu(), opSTMDAu(), opSTMDBu();

        static OPCODE opLDMIAw(), opLDMIBw(), opLDMDAw(), opLDMDBw();
        static OPCODE opSTMIAw(), opSTMIBw(), opSTMDAw(), opSTMDBw();

        static OPCODE opLDMIAuw(), opLDMIBuw(), opLDMDAuw(), opLDMDBuw();
        static OPCODE opSTMIAuw(), opSTMIBuw(), opSTMDAuw(), opSTMDBuw();

        //---SoftWare Interrupts---------------------------------------------------

        static OPCODE opBKPT(), opSWI();

        //---Coprocessor handlers--------------------------------------------------

	static OPCODE opMRC(), opMCR(), opCDP();
	static OPCODE opLDCofm(), opLDCofp();
	static OPCODE opLDCprm(), opLDCprp();
	static OPCODE opLDCptm(), opLDCptp();
	static OPCODE opLDCunm(), opLDCunp();
	static OPCODE opSTCofm(), opSTCofp();
	static OPCODE opSTCprm(), opSTCprp();
	static OPCODE opSTCptm(), opSTCptp();
	static OPCODE opSTCunm(), opSTCunp();
	
        //---eDSP extension--------------------------------------------------------
	
        static OPCODE opLDRDptim(), opLDRDofim(), opLDRDprim();
        static OPCODE opLDRDptip(), opLDRDofip(), opLDRDprip();
        static OPCODE opLDRDptrm(), opLDRDofrm(), opLDRDprrm();
        static OPCODE opLDRDptrp(), opLDRDofrp(), opLDRDprrp();
        static OPCODE opSTRDptim(), opSTRDofim(), opSTRDprim();
        static OPCODE opSTRDptip(), opSTRDofip(), opSTRDprip();
        static OPCODE opSTRDptrm(), opSTRDofrm(), opSTRDprrm();
        static OPCODE opSTRDptrp(), opSTRDofrp(), opSTRDprrp();
	
        static OPCODE opQADD(), opQSUB(), opQDADD(), opQDSUB();

	static OPCODE opSMLABB(), opSMLABT(), opSMLATB(), opSMLATT();
	static OPCODE opSMULBB(), opSMULBT(), opSMULTB(), opSMULTT();
	static OPCODE opSMLALBB(), opSMLALBT(), opSMLALTB(), opSMLALTT();
	static OPCODE opSMLAWB(), opSMLAWT(), opSMULWB(), opSMULWT();
	
        //---Thumb ops-------------------------------------------------------------

        static OPCODE topUND();

        static OPCODE topLSLimm(), topLSRimm(), topASRimm();
        static OPCODE topADDreg(), topSUBreg();
        static OPCODE topADDimm3(), topSUBimm3();

        static OPCODE topMOVimm8r0(), topMOVimm8r1();
        static OPCODE topMOVimm8r2(), topMOVimm8r3();
        static OPCODE topMOVimm8r4(), topMOVimm8r5();
        static OPCODE topMOVimm8r6(), topMOVimm8r7();
        static OPCODE topCMPimm8r0(), topCMPimm8r1();
        static OPCODE topCMPimm8r2(), topCMPimm8r3();
        static OPCODE topCMPimm8r4(), topCMPimm8r5();
        static OPCODE topCMPimm8r6(), topCMPimm8r7();
        static OPCODE topADDimm8r0(), topADDimm8r1();
        static OPCODE topADDimm8r2(), topADDimm8r3();
        static OPCODE topADDimm8r4(), topADDimm8r5();
        static OPCODE topADDimm8r6(), topADDimm8r7();
        static OPCODE topSUBimm8r0(), topSUBimm8r1();
        static OPCODE topSUBimm8r2(), topSUBimm8r3();
        static OPCODE topSUBimm8r4(), topSUBimm8r5();
        static OPCODE topSUBimm8r6(), topSUBimm8r7();

        static OPCODE topDPg1(), topDPg2(), topDPg3(), topDPg4();
        static OPCODE topADDH(), topMOVH(), topCMPH(), topBXreg();

        static OPCODE topLDRPCr0(), topLDRPCr1();
        static OPCODE topLDRPCr2(), topLDRPCr3();
        static OPCODE topLDRPCr4(), topLDRPCr5();
        static OPCODE topLDRPCr6(), topLDRPCr7();

        static OPCODE topSTRSPr0(), topSTRSPr1();
        static OPCODE topSTRSPr2(), topSTRSPr3();
        static OPCODE topSTRSPr4(), topSTRSPr5();
        static OPCODE topSTRSPr6(), topSTRSPr7();
        static OPCODE topLDRSPr0(), topLDRSPr1();
        static OPCODE topLDRSPr2(), topLDRSPr3();
        static OPCODE topLDRSPr4(), topLDRSPr5();
        static OPCODE topLDRSPr6(), topLDRSPr7();

        static OPCODE topSTRreg(),  topLDRreg();
        static OPCODE topSTRHreg(), topLDRHreg();
        static OPCODE topSTRBreg(), topLDRBreg();

        static OPCODE topLDRSBreg(), topLDRSHreg();

        static OPCODE topSTRimm5(),  topLDRimm5();
        static OPCODE topSTRBimm5(), topLDRBimm5();
        static OPCODE topSTRHimm5(), topLDRHimm5();

        static OPCODE topADDPCr0(), topADDPCr1();
        static OPCODE topADDPCr2(), topADDPCr3();
        static OPCODE topADDPCr4(), topADDPCr5();
        static OPCODE topADDPCr6(), topADDPCr7();
        static OPCODE topADDSPr0(), topADDSPr1();
        static OPCODE topADDSPr2(), topADDSPr3();
        static OPCODE topADDSPr4(), topADDSPr5();
        static OPCODE topADDSPr6(), topADDSPr7();
        static OPCODE topADDSPimm7();

        static OPCODE topPUSH(), topPUSHlr();
        static OPCODE topPOP(),  topPOPpc();

        static OPCODE topBKPT(), topSWI();

        static OPCODE topLDMIAr0(), topLDMIAr1();
        static OPCODE topLDMIAr2(), topLDMIAr3();
        static OPCODE topLDMIAr4(), topLDMIAr5();
        static OPCODE topLDMIAr6(), topLDMIAr7();
        static OPCODE topSTMIAr0(), topSTMIAr1();
        static OPCODE topSTMIAr2(), topSTMIAr3();
        static OPCODE topSTMIAr4(), topSTMIAr5();
        static OPCODE topSTMIAr6(), topSTMIAr7();

        static OPCODE topBEQ(), topBNE(), topBCS(), topBCC();
        static OPCODE topBMI(), topBPL(), topBVS(), topBVC();
        static OPCODE topBHI(), topBLS(), topBGE(), topBLT();
        static OPCODE topBGT(), topBLE();

        static OPCODE topB(), topBLsetup(), topBLoff(), topBLXoff();
};

// Allow for cross-compilation of the common opcode implementations
#define OPC ARM9ES::OPCODE ARM9ES::

#endif//__ARM9ES_H_

/*** EOF: arm9es.h *******************************************************/

