/**************************************************************************
* DSemu - The Next Generation                                             *
* GBA memory management: Plugin implementation [gbammu.cpp]               *
* Copyright Imran Nazar, 2005; released under the BSD public licence.     *
**************************************************************************/

#include <string>
#include <fstream>
#include "defs.h"
#include "dsmmusub.h"
#include "arm7tdmi.h"
#include "plgcpu.h"
#include "config.h"
#include "log.h"
#include "err.h"
#include "font5x7.h"
#include "byteswap.h"

//---Static private class members------------------------------------------
// Every plugin has an INFO structure attached, with info about the plugin.

PLUGININFO dsMMUsub::pInfo={
    PLUGIN_TYPE_MMU,
    0x00010001,
    "DS SubCPU memory manager",
    "DSemu-ng"
};

u8  *dsMMUsub::EWRAM , *dsMMUsub::IWRAM , *dsMMUsub::BIOS , *dsMMUsub::ROM ;
u16 *dsMMUsub::EWRAMh, *dsMMUsub::IWRAMh, *dsMMUsub::BIOSh, *dsMMUsub::ROMh;
u32 *dsMMUsub::EWRAMw, *dsMMUsub::IWRAMw, *dsMMUsub::BIOSw, *dsMMUsub::ROMw;

u8 *dsMMUsub::SWRAM; u16 *dsMMUsub::SWRAMh; u32 *dsMMUsub::SWRAMw;

std::string dsMMUsub::pluginName;

GUIPlugin *dsMMUsub::GUI=NULL;
ARM7TDMI *dsMMUsub::CPU=NULL;

int dsMMUsub::dmpwinID;
u32 *dsMMUsub::dmpbuffer;
u32 dsMMUsub::dmpaddr = 0x04000000;

dsMMUsub::PAGE dsMMUsub::mmioPages[256];
dsMMUsub::IOREG dsMMUsub::interruptIO[4];

u32 dsMMUsub::ROMsize;
u32 dsMMUsub::BIOSsize;
std::string dsMMUsub::ROMfile;

int dsMMUsub::waitSRAM, dsMMUsub::waitROM0;
int dsMMUsub::waitROM1, dsMMUsub::waitROM2;
int dsMMUsub::waitROM0s, dsMMUsub::waitROM1s, dsMMUsub::waitROM2s;
u32 dsMMUsub::prevROMaddr;

// Pointer to the DMA subhandler
dsMMUsub::DMA *dsMMUsub::dma=NULL;

//---Implementation--------------------------------------------------------

// Load a binary
void dsMMUsub::load(std::string fname)
{
    ROMfile = std::string(fname);
	
    NDShead hd(fname);
    
    std::ifstream input(fname.c_str(), std::ios::binary);
    u8 *tmp = new u8[hd.head.arm7size];
    if(!tmp) throw Exception(ERR_MMU_LOAD, pName,
		             "Couldn't allocate temporary ROM space.");

    input.seekg(hd.head.arm7src, std::ios::beg);
    input.read((char*)tmp, hd.head.arm7size);
    input.close();
    
    ROMsize=nlpo2(hd.head.arm7size);
    for(u32 i=0; i<hd.head.arm7size>>2; i++)
	wrW(hd.head.arm7dest+(i<<2), ((u32*)tmp)[i]);
    
    CPU->setPC(hd.head.arm7exec);

    char str[256];
    sprintf(str, "Loaded %d bytes to %08X, starting ARM7 execution at %08X.",
            hd.head.arm7size, hd.head.arm7dest, hd.head.arm7exec);
    Logger::log(pluginName) << str;
}

// Switch privilege levels (totally ignored in GBA's MMU)
void dsMMUsub::priv(u8 level) { }

void dsMMUsub::waitstates(u16 r)
{
    switch(r&0x0003)
    {
	case 0: waitSRAM=4; break; case 1: waitSRAM=3; break;
	case 2: waitSRAM=2; break; case 3: waitSRAM=8; break;
    }
    switch(r&0x000C)
    {
	case 0x0000: waitROM0=4; break; case 0x0004: waitROM0=3; break;
	case 0x0008: waitROM0=2; break; case 0x000C: waitROM0=8; break;
    }
    waitROM0s=(r&0x0010)?1:2;
    switch(r&0x0060)
    {
	case 0x0000: waitROM1=4; break; case 0x0020: waitROM1=3; break;
	case 0x0040: waitROM1=2; break; case 0x0060: waitROM1=8; break;
    }
    waitROM1s=(r&0x0080)?1:4;
    switch(r&0x0300)
    {
	case 0x0000: waitROM2=4; break; case 0x0100: waitROM2=3; break;
	case 0x0200: waitROM2=2; break; case 0x0300: waitROM2=8; break;
    }
    waitROM2s=(r&0x0400)?1:8;
    prevROMaddr=0;
}

// NOTE: The MMU works by splitting the address space into pages, each of
// which may be assigned a set of access handlers. The default handlers
// simply signal an unknown access; these are overwritten if a plugin
// wants to do something interesting.

// Register a range (set the pagetable entry)
void dsMMUsub::rangeReg(u8 page, rdBptr _rdB, rdHptr _rdH, rdWptr _rdW,
                               wrBptr _wrB, wrHptr _wrH, wrWptr _wrW)
{
    if(!pagetable[page].set)
    {
        if(_rdB) pagetable[page].rdB=_rdB;
        if(_rdH) pagetable[page].rdH=_rdH;
        if(_rdW) pagetable[page].rdW=_rdW;
        if(_wrB) pagetable[page].wrB=_wrB;
        if(_wrH) pagetable[page].wrH=_wrH;
        if(_wrW) pagetable[page].wrW=_wrW;
	pagetable[page].set = 1;
    }
}

// Wrappers for the pagetable entry for a given address
// NOTE: Endianness is automatically adjusted between host and emulation
u8 dsMMUsub::rdB(u32 addr) { return pagetable[PAGE_INDEX].rdB(addr); }
u16 dsMMUsub::rdH(u32 addr){ return mtohs(pagetable[PAGE_INDEX].rdH(addr)); }
u32 dsMMUsub::rdW(u32 addr){ return mtohl(pagetable[PAGE_INDEX].rdW(addr)); }

void dsMMUsub::wrB(u32 addr, u8 data) { pagetable[PAGE_INDEX].wrB(addr, data); }
void dsMMUsub::wrH(u32 addr, u16 data){ pagetable[PAGE_INDEX].wrH(addr, htoms(data)); }
void dsMMUsub::wrW(u32 addr, u32 data){ pagetable[PAGE_INDEX].wrW(addr, htoml(data)); }

// Region-specific I/O
// NOTE: The BIOS and ROM are (of course) non-writable
u8  dsMMUsub::rdB_EWRAM(u32 addr) { CPU->clockAdd(2); return EWRAM [(addr&0x003FFFFF)   ]; }
u16 dsMMUsub::rdH_EWRAM(u32 addr) { CPU->clockAdd(2); return EWRAMh[(addr&0x003FFFFF)>>1]; }
u32 dsMMUsub::rdW_EWRAM(u32 addr) { CPU->clockAdd(5); return EWRAMw[(addr&0x003FFFFF)>>2]; }
u8  dsMMUsub::rdB_IWRAM(u32 addr) { return IWRAM [(addr&0x0000FFFF)   ]; }
u16 dsMMUsub::rdH_IWRAM(u32 addr) { return IWRAMh[(addr&0x0000FFFF)>>1]; }
u32 dsMMUsub::rdW_IWRAM(u32 addr) { return IWRAMw[(addr&0x0000FFFF)>>2]; }
u8  dsMMUsub::rdB_SWRAM(u32 addr) { return SWRAM [(addr&0x0000FFFF)   ]; }
u16 dsMMUsub::rdH_SWRAM(u32 addr) { return SWRAMh[(addr&0x0000FFFF)>>1]; }
u32 dsMMUsub::rdW_SWRAM(u32 addr) { return SWRAMw[(addr&0x0000FFFF)>>2]; }
u8  dsMMUsub::rdB_BIOS (u32 addr) { return BIOS  [(addr&BIOSsize  )   ]; }
u16 dsMMUsub::rdH_BIOS (u32 addr) { return BIOSh [(addr&BIOSsize  )>>1]; }
u32 dsMMUsub::rdW_BIOS (u32 addr) { return BIOSw [(addr&BIOSsize  )>>2]; }

u8  dsMMUsub::rdB_ROM  (u32 addr)
{
    switch(addr&0x0E000000)
    {
	case 0x08000000: CPU->clockAdd((addr==(prevROMaddr+1))?waitROM0s:waitROM0); break;
	case 0x0A000000: CPU->clockAdd((addr==(prevROMaddr+1))?waitROM1s:waitROM1); break;
	case 0x0C000000: CPU->clockAdd((addr==(prevROMaddr+1))?waitROM2s:waitROM2); break;
    }
    prevROMaddr=addr;
    return ROM   [(addr&ROMsize   )   ];
}

u16 dsMMUsub::rdH_ROM  (u32 addr)
{
    switch(addr&0x0E000000)
    {
	case 0x08000000: CPU->clockAdd((addr==(prevROMaddr+2))?waitROM0s:waitROM0); break;
	case 0x0A000000: CPU->clockAdd((addr==(prevROMaddr+2))?waitROM1s:waitROM1); break;
	case 0x0C000000: CPU->clockAdd((addr==(prevROMaddr+2))?waitROM2s:waitROM2); break;
    }
    prevROMaddr=addr;
    return ROMh  [(addr&ROMsize   )>>1];
}

u32 dsMMUsub::rdW_ROM  (u32 addr)
{
    switch(addr&0x0E000000)
    {
	case 0x08000000: CPU->clockAdd((addr==(prevROMaddr+4))?waitROM0s+3:waitROM0+3); break;
	case 0x0A000000: CPU->clockAdd((addr==(prevROMaddr+4))?waitROM1s+3:waitROM1+3); break;
	case 0x0C000000: CPU->clockAdd((addr==(prevROMaddr+4))?waitROM2s+3:waitROM2+3); break;
    }
    prevROMaddr=addr;
    return ROMw  [(addr&ROMsize   )>>2];
}

void dsMMUsub::wrB_EWRAM(u32 a, u8  d) { CPU->clockAdd(2); EWRAM [(a&0x003FFFFF)   ]=d; }
void dsMMUsub::wrH_EWRAM(u32 a, u16 d) { CPU->clockAdd(2); EWRAMh[(a&0x003FFFFF)>>1]=d; }
void dsMMUsub::wrW_EWRAM(u32 a, u32 d) { CPU->clockAdd(5); EWRAMw[(a&0x003FFFFF)>>2]=d; }
void dsMMUsub::wrB_IWRAM(u32 a, u8  d) { IWRAM [(a&0x0000FFFF)   ]=d; }
void dsMMUsub::wrH_IWRAM(u32 a, u16 d) { IWRAMh[(a&0x0000FFFF)>>1]=d; }
void dsMMUsub::wrW_IWRAM(u32 a, u32 d) { IWRAMw[(a&0x0000FFFF)>>2]=d; }
void dsMMUsub::wrB_SWRAM(u32 a, u8  d) { SWRAM [(a&0x00007FFF)   ]=d; }
void dsMMUsub::wrH_SWRAM(u32 a, u16 d) { SWRAMh[(a&0x00007FFF)>>1]=d; }
void dsMMUsub::wrW_SWRAM(u32 a, u32 d) { SWRAMw[(a&0x00007FFF)>>2]=d; }
void dsMMUsub::wrB_BIOS (u32 a, u8  d) { }
void dsMMUsub::wrH_BIOS (u32 a, u16 d) { }
void dsMMUsub::wrW_BIOS (u32 a, u32 d) { }
void dsMMUsub::wrB_ROM  (u32 a, u8  d) { }
void dsMMUsub::wrH_ROM  (u32 a, u16 d) { }
void dsMMUsub::wrW_ROM  (u32 a, u32 d) { }

// Default pagetable entries (signal that something bad happened)
u8 dsMMUsub::rdB_BAD(u32 a)
{
    char str[128];
    sprintf(str, "Bad access: rdB %08X", a);
//    Logger::log(pluginName) << str;
    return 0;
}

u16 dsMMUsub::rdH_BAD(u32 a)
{
    char str[128];
    sprintf(str, "Bad access: rdH %08X", a);
//    Logger::log(pluginName) << str;
    return 0;
}

u32 dsMMUsub::rdW_BAD(u32 a)
{
    char str[128];
    sprintf(str, "Bad access: rdW %08X", a);
//    Logger::log(pluginName) << str;
    return 0;
}

void dsMMUsub::wrB_BAD(u32 a, u8 d)
{
    char str[128];
    sprintf(str, "Bad access: wrB %08X, %02X", a, d);
    Logger::log(pluginName) << str;
}

void dsMMUsub::wrH_BAD(u32 a, u16 d)
{
    char str[128];
    sprintf(str, "Bad access: wrH %08X, %04X", a, d);
    Logger::log(pluginName) << str;
}

void dsMMUsub::wrW_BAD(u32 a, u32 d)
{
    char str[128];
    sprintf(str, "Bad access: wrW %08X, %08X", a, d);
    Logger::log(pluginName) << str;
}

// Memory-mapped I/O registration. This works similarly to main memory,
// by paging the space. Of course, each page is much smaller.

void dsMMUsub::mmioReg(u8 page, rdBptr _rdB, rdHptr _rdH, rdWptr _rdW,
                              wrBptr _wrB, wrHptr _wrH, wrWptr _wrW)
{
    if(!mmioPages[page].set)
    {
        if(_rdB) mmioPages[page].rdB=_rdB;
        if(_rdH) mmioPages[page].rdH=_rdH;
        if(_rdW) mmioPages[page].rdW=_rdW;
        if(_wrB) mmioPages[page].wrB=_wrB;
        if(_wrH) mmioPages[page].wrH=_wrH;
        if(_wrW) mmioPages[page].wrW=_wrW;
	mmioPages[page].set = 1;
    }
}

// Wrapper functions that will be called to access MMIO; these index
// the MMIO page table.
// NOTE: No endian conversions are done here; they're performed by the
// top-level rd?/wr? functions.

u8 dsMMUsub::rdB_MMIO(u32 a) { return mmioPages[MMIO_PAGE_INDEX].rdB(a); }
u16 dsMMUsub::rdH_MMIO(u32 a){ return mmioPages[MMIO_PAGE_INDEX].rdH(a); }
u32 dsMMUsub::rdW_MMIO(u32 a){ return mmioPages[MMIO_PAGE_INDEX].rdW(a); }

void dsMMUsub::wrB_MMIO(u32 a, u8 d) { mmioPages[MMIO_PAGE_INDEX].wrB(a,d); }
void dsMMUsub::wrH_MMIO(u32 a, u16 d){ mmioPages[MMIO_PAGE_INDEX].wrH(a,d); }
void dsMMUsub::wrW_MMIO(u32 a, u32 d){ mmioPages[MMIO_PAGE_INDEX].wrW(a,d); }

// Since there's no other real place to keep the BIOS interrupt flags,
// they're here in the MMU, along with their accessor functions.
u8 dsMMUsub::rdB_INT(u32 a) { return interruptIO[(a&15)>>2].b[a&3]; }
u16 dsMMUsub::rdH_INT(u32 a){ return interruptIO[(a&15)>>2].h[(a&2)>>1]; }
u32 dsMMUsub::rdW_INT(u32 a){ return interruptIO[(a&15)>>2].w; }

void dsMMUsub::wrB_INT(u32 a, u8 d)
{
    interruptIO[(a&15)>>2].b[a&3]=d;
    switch(a&15)
    {
	case 4: waitstates((interruptIO[1].b[1]<<8)|d); break;
	case 5: waitstates(interruptIO[1].b[0]|(d<<8)); break;
    }
}
void dsMMUsub::wrH_INT(u32 a, u16 d){ interruptIO[(a&15)>>2].h[(a&2)>>1]=d; if((a&14)==4) waitstates(d); }
void dsMMUsub::wrW_INT(u32 a, u32 d){ interruptIO[(a&15)>>2].w=d; waitstates(d&65535); }

// If the MMU is to be informed of events, this is where they come in.
// At present, this merely wraps the DMA event handler.
void dsMMUsub::event(int type, void *data)
{
    switch(type)
    {
	case MMU_EVENT_SETEWRAM:
	    EWRAM =(u8 *)data;
	    EWRAMh=(u16*)data;
	    EWRAMw=(u32*)data;
	    for(int i=0;i<16;i++)
                rangeReg(0x20+i, rdB_EWRAM, rdH_EWRAM, rdW_EWRAM, wrB_EWRAM, wrH_EWRAM, wrW_EWRAM);
	    break;

	case MMU_EVENT_SETSWRAM:
	    SWRAM =(u8 *)data;
	    SWRAMh=(u16*)data;
	    SWRAMw=(u32*)data;
            rangeReg(0x37, rdB_SWRAM, rdH_SWRAM, rdW_SWRAM, wrB_SWRAM, wrH_SWRAM, wrW_SWRAM);
	    break;

        default: dma->check(type); break;
    }
}

// Initialise plugin
// Parameters: name  - FQPN of plugin as listed in INI file
//             req   - Pointer to PluginRequest API function
//             unreq - Pointer to PluginUnrequest API function
dsMMUsub::dsMMUsub(std::string name, REQPTR req, UNREQPTR unreq)
{
    pName = std::string(name);
    pClass = pName.substr(0, pName.find(".")+1);
    pRequest = req;
    pUnrequest = unreq;

    pluginName = std::string(pName);
    
    ROMfile = std::string("");
    if(!GUI) GUI = (GUIPlugin*)pRequest("UI");

    dmpbuffer = new u32[360*128];
    dmpwinID = GUI->subwinCreate(360,128, "Sub Memory Viewer", dmpbuffer);
    
    // Blank out the page table
    for(int i=0; i<256; ++i)
    {
        pagetable[i].rdB=rdB_BAD;
        pagetable[i].rdH=rdH_BAD;
        pagetable[i].rdW=rdW_BAD;
        pagetable[i].wrB=wrB_BAD;
        pagetable[i].wrH=wrH_BAD;
        pagetable[i].wrW=wrW_BAD;
	pagetable[i].set=0;

        mmioPages[i].rdB=rdB_BAD;
        mmioPages[i].rdH=rdH_BAD;
        mmioPages[i].rdW=rdW_BAD;
        mmioPages[i].wrB=wrB_BAD;
        mmioPages[i].wrH=wrH_BAD;
        mmioPages[i].wrW=wrW_BAD;
	mmioPages[i].set=0;
    }
    
    // Make all memory pointers null
    EWRAM=NULL; EWRAMh=NULL; EWRAMw=NULL;
    IWRAM=NULL; IWRAMh=NULL; IWRAMw=NULL;
    SWRAM=NULL; SWRAMh=NULL; SWRAMw=NULL;
    BIOS =NULL; BIOSh =NULL; BIOSw =NULL;
    ROM  =NULL; ROMh  =NULL; ROMw  =NULL;
    
    // Allocate RAMs
    /*
    EWRAM = new u8[256*1024];
    if(!EWRAM)
        throw Exception(ERR_MMU_INIT, pName,
                        "Allocation of EWRAM failed.");

    EWRAMh = (u16*)EWRAM;
    EWRAMw = (u32*)EWRAM;

    Logger::log(pName) << "256KB of EWRAM allocated.";
    */
    IWRAM = new u8[64*1024];
    if(!IWRAM)
        throw Exception(ERR_MMU_INIT, pName,
                        "Allocation of IWRAM failed.");
    IWRAMh = (u16*)IWRAM;
    IWRAMw = (u32*)IWRAM;

    Logger::log(pName) << "64KB of IWRAM allocated.";
    
    // Register regions with pagetable
    for(int i=0; i<16; ++i)
    {
        rangeReg(0x00+i, rdB_BIOS , rdH_BIOS , rdW_BIOS , wrB_BAD  , wrH_BAD  , wrW_BAD  );
    }
    for(int i=0; i<8; i++)
    {
        rangeReg(0x38+i, rdB_IWRAM, rdH_IWRAM, rdW_IWRAM, wrB_IWRAM, wrH_IWRAM, wrW_IWRAM);
        rangeReg(0x40+i, rdB_MMIO,  rdH_MMIO,  rdW_MMIO,  wrB_MMIO,  wrH_MMIO,  wrW_MMIO );
    }

    // Register the interrupt flag I/O space with the MMIO pagetable
    mmioReg(0x20, rdB_INT, rdH_INT, rdW_INT, wrB_INT, wrH_INT, wrW_INT);
    for(int i=0; i<4; ++i) interruptIO[i].w=0;
    waitstates(0);

    // All done.
    Logger::log(pName) << "Initialised.";
}

// Plugin cleanup
dsMMUsub::~dsMMUsub()
{
    //if(EWRAM != NULL) { delete EWRAM; EWRAM=NULL; EWRAMh=NULL; EWRAMw=NULL; }
    if(IWRAM != NULL) { delete IWRAM; IWRAM=NULL; IWRAMh=NULL; IWRAMw=NULL; }
    if(ROM   != NULL) { delete ROM;   ROM  =NULL; ROMh  =NULL; ROMw  =NULL; }
   
    if(dma) { delete dma; dma=NULL; }
    if(dmpbuffer) { delete dmpbuffer; dmpbuffer=NULL; }
    
    pUnrequest("UI", 0); GUI = NULL;
    
    Logger::log(pName) << "Shutdown.";
}

void dsMMUsub::setCPU(CPUPlugin *c)
{
    if(c) CPU = (ARM7TDMI*)c;
}

// Reset plugin state
void dsMMUsub::reset()
{
    // Clear out memory
    memset(IWRAM, 0, 64*1024);
    if(ROMfile.length()) load(ROMfile);

    memset(dmpbuffer, 0, 360*128*4);

    if(!dma) dma = new DMA(this); 
    
    dma->reset();
    
    // Since the BIOS file may change between resets, loading of the BIOS
    // is done here in reset as opposed to the constructor
    if(BIOS  != NULL) { delete BIOS;  BIOS =NULL; BIOSh =NULL; BIOSw =NULL; }

    Config ini; ini.read("dsmmusub.ini");
    if(!ini.exists("BIOS")) ini["BIOS"] = "bioshack.img";

    std::ifstream input(ini["BIOS"].c_str(), std::ios::binary);

    input.seekg(0, std::ios::end);
    BIOSsize = input.tellg();
    input.seekg(0, std::ios::beg);

    BIOS = new u8[nlpo2(BIOSsize)];
    if(!BIOS)
        throw Exception(ERR_MMU_INIT, pName,
                        "Allocation of BIOS space failed.");
    BIOSh=(u16*)BIOS;
    BIOSw=(u32*)BIOS;
    
    input.read((char*)BIOS, BIOSsize);
    input.close();

    BIOSsize = nlpo2(BIOSsize)-1;
    
    Logger::log(pName) << ini["BIOS"] << " (" << BIOSsize+1 <<" bytes) loaded as BIOS.";
    
    // Refresh the dump window
    status(0x04000000, 1);

    // We're done.
    Logger::log(pName) << "Reset.";
}

// Provide status information: fill the memorydump framebuffer
void dsMMUsub::status(int addr=0x04000000, int mode=1)
{
    Font5x7 prn(dmpbuffer, 360,128);
    char str[256];
    
    for(int i=0; i<16; ++i)
    {
	sprintf(str, "%08X | ", addr+i*16);
        switch(mode)
	{
	    case 0:
		for(int j=0; j<16; ++j) sprintf(str, "%s%02X ", str, rdB(addr+i*16+j));
		break;
	    case 1:
                for(int j=0; j<16; j+=2) sprintf(str, "%s%04X ", str, rdH(addr+i*16+j));
                break;
	    case 2:
		for(int j=0; j<16; j+=4) sprintf(str, "%s%08X ", str, rdW(addr+i*16+j));
		break;
	}
	prn.print(str,0,i*8,0x00FFFFFF);
    }

    GUI->subwinRefresh(dmpwinID);
}

// Information about what this MMU supports
// Returns: bits set in a u8
u8 dsMMUsub::getCaps()
{
    return PLGMMU_CAPS_32B | PLGMMU_CAPS_LSBEND;
}

// Next Largest Power of 2 (32-bit)
// This is used by the allocators to get nice chunks of memory, and make
// our masking life easy.
// ACK: Parallel Computing Aggregate, Kentucky U (nlpo2)

inline unsigned int
dsMMUsub::nlpo2(unsigned int x)
{
	if(x&(x-1)) {            // If we're not already a power of 2
	        x |= (x >> 1);
        	x |= (x >> 2);
	        x |= (x >> 4);
        	x |= (x >> 8);
	        x |= (x >> 16);
        	return(x+1);     // Return next power up
	}
	return x;                // Otherwise, stick with what we have
}

//---Plugin architecture support-------------------------------------------
// Retrieve Plugin class from outside
// Parameters: plg   - Address of a pointer to a Plugin class to 'new'
//             name  - FQPN of plugin as listed in INI file
//             req   - Pointer to PluginRequest API function
//             unreq - Pointer to PluginUnrequest API function
EXPORTFUNC void getPlugin(Plugin **plg,
                          std::string name,
                          REQPTR req, UNREQPTR unreq)
{
   *plg = new dsMMUsub(name, req, unreq);
}

// Provide plugin version information
PLUGININFO *dsMMUsub::getinfo()
{
    return &pInfo;
}

// Release plugin from outside
void dsMMUsub::release()
{
    // Delete the Test plugin that was 'new'd in getPlugin.
    delete this;
}

/*** EOF: gbammu.cpp *****************************************************/

