/**************************************************************************
* DSemu - The Next Generation                                             *
* GBA scanline GPU: Plugin Implementation [gbagpu.cpp]                    *
* Copyright Imran Nazar, 2005; released under the BSD public licence.     *
**************************************************************************/

#include <string>
#include "defs.h"
#include "events.h"
#include "gbagpu.h"
#include "config.h"
#include "font5x7.h"
#include "log.h"
#include "err.h"

//---Static private class members------------------------------------------
// Every plugin has an INFO structure attached, with info about the plugin.

PLUGININFO gbaGPU::pInfo={
    PLUGIN_TYPE_GPU,
    0x00010001,
    "GBA graphics processor",
    "DSemu-ng"
};

u8  *gbaGPU::VRAMb, *gbaGPU::OAMb;
u16 *gbaGPU::VRAMh, *gbaGPU::PALh, *gbaGPU::OAMh;
u32 *gbaGPU::VRAMw, *gbaGPU::PALw, *gbaGPU::OAMw;

gbaGPU::IOREG gbaGPU::reg[0x30];

// IO access flags: 1=Readable, 2=Writable
int gbaGPU::rflags[0x30]={
    3,3,3,1,3,3,3,3,3,3,3,3,3,3,3,3,
    3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
    3,3,3,3,3,3,3,0,3,3,3,0,0,0,0,0,
};

#define REG(x) reg[REG_ ## x].data

u32 gbaGPU::BGX[2], gbaGPU::BGY[2];
u8 gbaGPU::mospy[4];

int gbaGPU::curline;
std::string gbaGPU::pluginName;

GUIPlugin *gbaGPU::GUI;
int gbaGPU::palwinID;
u32 *gbaGPU::palbuffer;

u8 *gbaGPU::dispbuffer;

int gbaGPU::ForceBG[4];

// Convert a GBA 15-bit colour to 32-bit
#define COLSHIFT(out, in) \
{ \
    *((out)+0) = ((in)<<3)&255; \
    *((out)+1) = ((in)>>2)&255; \
    *((out)+2) = ((in)>>7)&255; \
    *((out)+3) = 0; \
}

// Table of function pointers to mode-specific renderers
gbaGPU::modefptr gbaGPU::lineM[8]={
    lineM0, lineM1, lineM2, lineM3,
    lineM4, lineM5, lineMNULL, lineMNULL
};

//---Implementation--------------------------------------------------------
// ACK: gladius (tile and rotated BG, and sprite renderers)

void gbaGPU::scanTile(u8 *buf, int bg)
{
    u16 bgcnt, scrnBase, charBase, hofs, vofs, x, y, tileIdx, tileY;
    u16 tmpTileIdx, width=0, height=0, a, bgx, bgy, tileChar; u8 b=0, palnum;
    u8 mosx, mospx, mosy;
    bgcnt=reg[REG_BG0CNT+bg].data;
    switch((bgcnt>>14)&3)
    {
        case 0: width=256; height=256; break;
        case 1: width=512; height=256; break;
        case 2: width=256; height=512; break;
        case 3: width=512; height=512; break;
    }
    scrnBase=((bgcnt>>8)&0x1F)*0x400;
    charBase=((bgcnt>>2)&0x3)*0x4000;
    hofs=reg[REG_BG0HOFS+bg*2].data;
    vofs=reg[REG_BG0VOFS+bg*2].data;
/*    vramptr=VRAM+(VRAMmap[(lcd)*128]<<13);
    vram8ptr=VRAM8+(VRAMmap[(lcd)*128]<<14);*/
    if(bgcnt&0x0040)
    {
        mosx=(REG(MOSAIC)&0x000F);
        mosy=((REG(MOSAIC)&0x00F0)>>4);
        mospx=mosx; if(!curline) mospy[bg]=mosy;
        mospy[bg]++; if(mospy[bg]<=mosy) vofs-=mospy[bg];
        else mospy[bg]=0;
    } else { mospx=0; mosx=0; }
    if(bgcnt&0x0080)
    {
        bgy=((curline+vofs)&(height-1))/8;
        tileIdx=scrnBase+(((bgy&31)*32));
        switch((bgcnt>>14)&3)
        {
            case 2: if(bgy>=32) tileIdx += 32*32; break;
            case 3: if(bgy>=32) tileIdx += 32*32*2; break;
        }
        tileY=((curline+vofs)&7)*8;
        for(a=0;a<240;a++)
        {
            bgx=((a+hofs)&(width-1))/8;
            tmpTileIdx=tileIdx+((bgx&31));
            if(bgx>=32) tmpTileIdx += 32*32;
            tileChar=VRAMh[tmpTileIdx];
            x=(a+hofs)&7; y=tileY;
            if(tileChar&0x0400) x=7-x;
            if(tileChar&0x0800) y=56-y;
            if(mospx>=mosx) b=VRAMb[charBase+((tileChar&0x03FF)*64)+x+y];
            mospx++; if(mospx>mosx) mospx=0;
            if(b) COLSHIFT(buf+a*4,PALh[b]);
        }
    } else {
        bgy=((curline+vofs)&(height-1))/8;
        tileIdx=scrnBase+(((bgy&31)*32));
        switch((bgcnt>>14)&3)
        {
            case 2: if(bgy>=32) tileIdx += 32*32; break;
            case 3: if(bgy>=32) tileIdx += 32*32*2; break;
        }
        tileY=((curline+vofs)&7)*4;
        for(a=0;a<240;a++)
        {
            bgx=((a+hofs)&(width-1))/8;
            tmpTileIdx=tileIdx+((bgx&31));
            if(bgx>=32) tmpTileIdx += 32*32;
            tileChar=VRAMh[tmpTileIdx];
            x=(a+hofs)&7; y=tileY;
            if(tileChar&0x0400) x=7-x;
            if(tileChar&0x0800) y=28-y;
            if(mospx>=mosx)
            {
                b=VRAMb[charBase+((tileChar&0x03FF)*32)+(x/2)+y];
                if(x&1) b>>=4; b&=15;
            }
            palnum=((tileChar>>12)&15)*16;
            mospx++; if(mospx>mosx) mospx=0;
            if(b) COLSHIFT(buf+a*4,PALh[b+palnum]);
        }
    }
}

void gbaGPU::scanRot(u8 *buf, int bg)
{
    u16 bgcnt, scrnBase, charBase;
    u32 x,y; int ax,ay; s16 dx,dy; int trans; u8 mosx, mospx, mosy;
    u16 tmpTileIdx, width=0, height=0, a, tileChar; u8 b=0;
    
    bgcnt=REG(BG0CNT+bg); bg-=2;
    trans=(bgcnt&0x2000)==0;
    switch((bgcnt>>14)&3)
    {
        case 0: width=128; height=128; break;
        case 1: width=256; height=256; break;
        case 2: width=512; height=512; break;
        case 3: width=1024; height=1024; break;
    }
    scrnBase=((bgcnt>>8)&0x1F)*0x800;
    charBase=((bgcnt>>2)&0x3)*0x4000;
    x=BGX[bg]; y=BGY[bg];
    dx=REG(BG2PA+(bg*8)); dy=REG(BG2PC+(bg*8));
    if(bgcnt&0x0040)
    {
        mosx=(REG(MOSAIC)&0x000F);
        mosy=((REG(MOSAIC)&0x00F0)>>4);
        mospx=mosx; if(!curline) mospy[bg]=mosy;
        mospy[bg]++; if(mospy[bg]<=mosy) y-=mospy[bg]*256;
        else mospy[bg]=0;
    } else { mospx=0; mosx=0; }
    for(a=0;a<240;a++)
    {
        ax=x>>8; ay=y>>8;
        if ((ax >= 0 && ax < width && ay >= 0 && ay < height) || !trans)
        {
            tmpTileIdx=scrnBase+(((ay&(height-1))/8)*(width/8)+((ax&(width-1))/8));
            tileChar=VRAMb[tmpTileIdx];
            if(mospx>=mosx) b=VRAMb[charBase+tileChar*64+((ay&7)*8)+(ax&7)];
            mospx+=(dx>>8); if(mospx>mosx) mospx=0;
            if(b) COLSHIFT(buf+a*4,PALh[b]);
        }
        x+=dx; y+=dy;
    }
}

void gbaGPU::sprites(u8 *buf, int pri)
{
    u16 dispcnt=REG(DISPCNT), oamNum, attr0,attr1,attr2;
    int x, y, i, width=-1, height=-1, rwidth, rheight, scale=1;
    int spritey, baseSprite, baseInc, curIdx, palIdx; u8 b;
    s16 dx,dmx,rx,dy,dmy,ry; int tx,ty,cx,cy,pitch,rotScaleParam;

    if(!(dispcnt&0x1000)) return;
    pri<<=10;
    oamNum=128; do
    {
	width=0; height=0;
        oamNum--;
        attr2=OAMh[oamNum*4+2];
        if((attr2&0x0C00)!=pri) continue;
        attr0=OAMh[oamNum*4+0];
        attr1=OAMh[oamNum*4+1];
        x=attr1&0x01FF; y=attr0&0x00FF;

                switch ((attr0 >> 14) & 3)
                {
                    case 0:
                        // Square
                        switch ((attr1 >> 14) & 3)
                        {
                            case 0: width = 8; height = 8; break;
                            case 1: width = 16; height = 16; break;
                            case 2: width = 32; height = 32; break;
                            case 3: width = 64; height = 64; break;
                        }
                        break;
                    case 1:
                        // Horizontal Rectangle
                        switch ((attr1 >> 14) & 3)
                        {
                            case 0: width = 16; height = 8; break;
                            case 1: width = 32; height = 8; break;
                            case 2: width = 32; height = 16; break;
                            case 3: width = 64; height = 32; break;
                        }
                        break;
                    case 2:
                        // Vertical Rectangle
                        switch ((attr1 >> 14) & 3)
                        {
                            case 0: width = 8; height = 16; break;
                            case 1: width = 8; height = 32; break;
                            case 2: width = 16; height = 32; break;
                            case 3: width = 32; height = 64; break;
                        }
                        break;
                }

                rwidth=width; rheight=height;
                if (attr0&0x0100)
                    // Rot-scale on
                    if (attr0&0x0200) { rwidth*=2; rheight*=2; }
                else
                    // Invalid sprite
                    if (attr0&0x0200) width = -1;

                if (width == -1)
                    // Invalid sprite
                    continue;

                // Y clipping
                if (y > ((y + rheight) & 0xff))
                    if (curline >= ((y + rheight) & 0xff) && !(y < curline)) continue;
                else
                    if (curline < y || curline >= ((y + rheight) & 0xff)) continue;

                if (attr0&0x2000) scale = 2;

                spritey = curline - y;
                if (spritey < 0) spritey += 256;

                if (!(attr0&0x0100))
                {
                    if (attr1&0x2000) spritey = (height - 1) - spritey;

                    if (dispcnt&0x0040)
                        // 1 dimensional
                        baseSprite = (attr2 & 0x3FF) + ((spritey / 8) * (width / 8)) * scale;
                    else
                        // 2 dimensional
                        baseSprite = (attr2 & 0x3FF) + ((spritey / 8) * 0x20);

                    baseInc = scale;
                    if (attr1&0x1000)
                    {
                        baseSprite += ((width / 8) * scale) - scale;
                        baseInc = -baseInc;
                    }

                    if (attr0&0x2000)
                    {
                        // 256 colors
			//if(baseSprite && spritey>=0 && spritey<height)
                        //printf("GPU: Line #%d: Sprite #%d: %dx%d at %d,%d, non-rotating, 8-bit\n",curline,oamNum,width,height,x,y);
                        for (i = x; i < x + width; i++)
                        {
                            if ((i & 0x1ff) < 240 && spritey >= 0 && spritey < height)
                            {
                                tx = (i - x) & 7;
                                if (attr1&0x1000) tx = 7 - tx;
                                curIdx = baseSprite * 32 + ((spritey & 7) * 8) + tx;
                                b = VRAMb[0x10000 + curIdx];
                                if (b) COLSHIFT(buf+(i&0x1ff)*4,PALh[b+256]);
                            }
                            if (((i - x) & 7) == 7) baseSprite += baseInc;
                        }
                    }
                    else
                    {
                        // 16 colors
			//if(baseSprite && spritey>=0 && spritey<height)
                        //printf("GPU: Line #%d: Sprite #%d: %dx%d at %d,%d, non-rotating, 4-bit\n",curline,oamNum,width,height,x,y);
                        palIdx = ((attr2 >> 8) & 0xF0);
                        for (i = x; i < x + width; i++)
                        {
                            if ((i & 0x1ff) < 240 && spritey >= 0 && spritey < height)
                            {
                                tx = (i - x) & 7;
                                if ((attr1 & (1 << 12)) != 0) tx = 7 - tx;
                                curIdx = baseSprite * 32 + ((spritey & 7) * 4) + (tx / 2);
                                b = VRAMb[0x10000 + curIdx];
                                if (tx & 1) b>>=4; b&=15;
                                if(b) COLSHIFT(buf+(i&0x1ff)*4, PALh[palIdx+b+256]);
                            }
                            if (((i - x) & 7) == 7) baseSprite += baseInc;
                        }
                    }
                }
                else
                {
                    rotScaleParam = (attr1 >> 9) & 0x1F;

                    dx =  (s16)OAMh[(rotScaleParam * 4 * 4) + 3];
                    dmx = (s16)OAMh[(rotScaleParam * 4 * 4) + 7];
                    dy =  (s16)OAMh[(rotScaleParam * 4 * 4) + 11];
                    dmy = (s16)OAMh[(rotScaleParam * 4 * 4) + 15];

                    cx = rwidth / 2;
                    cy = rheight / 2;

                    baseSprite = attr2 & 0x3FF;

                    if (dispcnt&0x0040)
                        // 1 dimensional
                        pitch = (width / 8) * scale;
                    else
                        // 2 dimensional
                        pitch = 0x20;

                    rx = (s16)((dmx * (spritey - cy)) - (cx * dx) + (width << 7));
                    ry = (s16)((dmy * (spritey - cy)) - (cx * dy) + (height << 7));

                    // Draw a rot/scale sprite
                    if (attr0&0x2000)
                    {
			//if(baseSprite && spritey>=0 && spritey<height)
                        //printf("GPU: Line #%d: Sprite #%d: %dx%d at %d,%d, rotating, 8-bit\n",curline,oamNum,width,height,x,y);
                        // 256 colors
                        for (i = x; i < x + rwidth; i++)
                        {
                            tx = rx >> 8;
                            ty = ry >> 8;

                            if ((i & 0x1ff) < 240 && tx >= 0 && tx < width && ty >= 0 && ty < height)
                            {
                                curIdx = (baseSprite + ((ty / 8) * pitch) + ((tx / 8) * scale)) * 32 + ((ty & 7) * 8) + (tx & 7);
                                b = VRAMb[0x10000 + curIdx];
                                if(b) COLSHIFT(buf+(i&0x1ff)*4, PALh[b+256]);
                            }

                            rx += dx;
                            ry += dy;
                        }
                    }
                    else
                    {
                        // 16 colors
			//if(baseSprite && spritey>=0 && spritey<height)
                        //printf("GPU: Line #%d: Sprite #%d: %dx%d at %d,%d, rotating, 4-bit\n",curline,oamNum,width,height,x,y);
                        palIdx = ((attr2 >> 8) & 0xF0);
                        for (i = x; i < x + rwidth; i++)
                        {
                            tx = rx >> 8;
                            ty = ry >> 8;

                            if ((i & 0x1ff) < 240 && tx >= 0 && tx < width && ty >= 0 && ty < height)
                            {
                                curIdx = (baseSprite + ((ty / 8) * pitch) + ((tx / 8) * scale)) * 32 + ((ty & 7) * 4) + ((tx & 7) / 2);
                                b = VRAMb[0x10000 + curIdx];
                                if (tx & 1) b>>=4; b&=15;
                                if(b) COLSHIFT(buf+(i&0x1ff)*4, PALh[palIdx+b+256]);
                            }

                            rx += dx;
                            ry += dy;
                        }
                    }
                }
    } while(oamNum);
}

// NOTE: Each renderer draws the backgrounds in reverse priority order,
// rendering the sprites on each priority level after the background.

// Mode 0: 4 tiled BGs
void gbaGPU::lineM0(u8 *buf)
{
    u16 dispcnt=REG(DISPCNT);
    for(int pri=3;pri>=0;pri--)
    {
        if(ForceBG[3]!=2) if(dispcnt&0x0800 || ForceBG[3]) if((REG(BG3CNT)&3)==pri) scanTile(buf,3);
        if(ForceBG[2]!=2) if(dispcnt&0x0400 || ForceBG[2]) if((REG(BG2CNT)&3)==pri) scanTile(buf,2);
        if(ForceBG[1]!=2) if(dispcnt&0x0200 || ForceBG[1]) if((REG(BG1CNT)&3)==pri) scanTile(buf,1);
        if(ForceBG[0]!=2) if(dispcnt&0x0100 || ForceBG[0]) if((REG(BG0CNT)&3)==pri) scanTile(buf,0);
        sprites(buf,pri);
    }
}

// Mode 1: Two tiled and 1 rot BG
void gbaGPU::lineM1(u8 *buf)
{
    u16 dispcnt=REG(DISPCNT);
    for(int pri=3;pri>=0;pri--)
    {
        if(ForceBG[3]!=2) if(dispcnt&0x0400 || ForceBG[2]) if((REG(BG2CNT)&3)==pri) scanRot(buf,2);
        if(ForceBG[2]!=2) if(dispcnt&0x0200 || ForceBG[1]) if((REG(BG1CNT)&3)==pri) scanTile(buf,1);
        if(ForceBG[1]!=2) if(dispcnt&0x0100 || ForceBG[0]) if((REG(BG0CNT)&3)==pri) scanTile(buf,0);
        sprites(buf,pri);
    }
}

// Mode 2: 2 rot backgrounds
void gbaGPU::lineM2(u8 *buf)
{
    u16 dispcnt=REG(DISPCNT);
    for(int pri=3;pri>=0;pri--)
    {
        if(ForceBG[3]!=2) if(dispcnt&0x0800 || ForceBG[3]) if((REG(BG3CNT)&3)==pri) scanRot(buf,3);
        if(ForceBG[2]!=2) if(dispcnt&0x0400 || ForceBG[2]) if((REG(BG2CNT)&3)==pri) scanRot(buf,2);
        sprites(buf,pri);
    }
}

// Mode 3: Flat (rotatable) 16-bit framebuffer
void gbaGPU::lineM3(u8 *buf)
{
    u16 dispcnt=REG(DISPCNT), bg2cnt=REG(BG2CNT), i; int pri;
    u32 x,y,ax,ay; s16 dx,dy; u16 mosx,mospx,mosy;
    for(pri=3;pri>(bg2cnt&3);pri--) sprites(buf,pri);
    if(ForceBG[2]!=2 && (dispcnt&0x0400 || ForceBG[2]==1))
    {
        x=BGX[0]; y=BGY[0];
        if(bg2cnt&0x0040)
        {
            mosx=(REG(MOSAIC)&0x000F);
            mosy=((REG(MOSAIC)&0x00F0)>>4);
            mospx=mosx; if(!curline) mospy[2]=mosy;
            mospy[2]++; if(mospy[2]<=mosy) y-=mospy[2]*256;
            else mospy[2]=0;
        } else { mospx=0; mosx=0; }
        dx=(signed)REG(BG2PA); dy=(signed)REG(BG2PC);
        for(i=0;i<240;i++)
        {
            ax=x>>8; ay=y>>8;
            if (ax >= 0 && ax < 240 && ay >= 0 && ay < 160)
            {
                if(mospx>=mosx) COLSHIFT(buf+i*4, VRAMh[ay*240+ax]);
            }
            mospx+=(dx>>8); if(mospx>mosx) mospx=0;
            x+=dx; y+=dy;
        }
    }
    for(pri=(bg2cnt&3);pri>=0;pri--) sprites(buf,pri);
}

// Mode 4: 2 paletted framebuffers, switchable on BG2CNT bit 4
void gbaGPU::lineM4(u8 *buf)
{
    u16 dispcnt=REG(DISPCNT), bg2cnt=REG(BG2CNT), i; int pri;
    u32 x,y,ax,ay; s16 dx,dy; u8 b=0; u16 mosx,mospx,mosy;
    u16 baseIdx=(bg2cnt&0x0010)?0xA000:0;
    for(pri=3;pri>(bg2cnt&3);pri--) sprites(buf,pri);
    if(ForceBG[2]!=2 && (dispcnt&0x0400 || ForceBG[2]==1))
    {
        x=BGX[0]; y=BGY[0];
        if(bg2cnt&0x0040)
        {
            mosx=(REG(MOSAIC)&0x000F);
            mosy=((REG(MOSAIC)&0x00F0)>>4);
            mospx=mosx; if(!curline) mospy[2]=mosy;
            mospy[2]++; if(mospy[2]<=mosy) y-=mospy[2]*256;
            else mospy[2]=0;
        } else { mospx=0; mosx=0; }
        dx=(signed)REG(BG2PA); dy=(signed)REG(BG2PC);
        for(i=0;i<240;i++)
        {
            ax=x>>8; ay=y>>8;
            if (ax >= 0 && ax < 240 && ay >= 0 && ay < 160)
            {
                if(mospx>=mosx) b=VRAMb[baseIdx+ay*240+ax];
                if(b) COLSHIFT(buf+i*4, PALh[b]);
            }
            mospx+=(dx>>8); if(mospx>mosx) mospx=0;
            x+=dx; y+=dy;
        }
    }
    for(pri=(bg2cnt&3);pri>=0;pri--) sprites(buf,pri);
}

// Mode 5: 2 reduced-size framebuffers, switchable on BG2CNT b4
void gbaGPU::lineM5(u8 *buf)
{
    u16 dispcnt=REG(DISPCNT), bg2cnt=REG(BG2CNT), i; int pri;
    u32 x,y,ax,ay; s16 dx,dy; u16 mosx,mospx,mosy;
    u16 baseIdx=(bg2cnt&0x0010)?(160*128):0;
    for(pri=3;pri>(bg2cnt&3);pri--) sprites(buf,pri);
    if(ForceBG[2]!=2 && (dispcnt&0x0400 || ForceBG[2]==1))
    {
        x=BGX[0]; y=BGY[0];
        if(bg2cnt&0x0040)
        {
            mosx=(REG(MOSAIC)&0x000F);
            mosy=((REG(MOSAIC)&0x00F0)>>4);
            mospx=mosx; if(!curline) mospy[2]=mosy;
            mospy[2]++; if(mospy[2]<=mosy) y-=mospy[2]*256;
            else mospy[2]=0;
        } else { mospx=0; mosx=0; }
        dx=(signed)REG(BG2PA); dy=(signed)REG(BG2PC);
        for(i=0;i<240;i++)
        {
            ax=x>>8; ay=y>>8;
            if (ax >= 0 && ax < 160 && ay >= 0 && ay < 128)
            {
                if(mospx>=mosx) COLSHIFT(buf+i*4, VRAMh[baseIdx+ay*160+ax]);
            }
            mospx+=(dx>>8); if(mospx>mosx) mospx=0;
            x+=dx; y+=dy;
        }
    }
    for(pri=(bg2cnt&3);pri>=0;pri--) sprites(buf,pri);
}

// Null renderer: Should never have to draw this.
void gbaGPU::lineMNULL(u8 *buf)
{
    Logger::log(pluginName) << "How'd we get to lineMNULL?";
}

// Tell the GPU where it'll be rendering to
void gbaGPU::setDisplay(u8 *buf)
{
    dispbuffer=buf;
}

// Render a scanline (wrapper for mode-specific renderers)
void gbaGPU::line()
{
    u16 dispcnt = REG(DISPCNT);
    u8 *buffer = dispbuffer;
    if(curline >= 0 && curline <= 160)
    {
	buffer += (curline*240*4);
	if(dispcnt&0x0080)  // Blank screen
	{
	    for(int i=0; i<240; ++i) COLSHIFT(buffer+i*4, 0x00FFFFFF);
        }
	else  // First draw the backdrop, then the line
	{
	    for(int i=0; i<240; ++i) COLSHIFT(buffer+i*4, PALh[0]);
	    lineM[dispcnt&7](buffer);
	}
    }

    // The line's done, so hblank begins
    HBLstart();
}

// Beginning of hblank: Set hbl bit, fire interrupt
void gbaGPU::HBLstart()
{
    reg[REG_DISPSTAT].b[0] |= 2;
    if(reg[REG_DISPSTAT].b[0] & 16)
    {
	if((MMU->rdH(0x04000200) & 2) && MMU->rdH(0x04000208))
	{
	    MMU->wrH(0x04000202, MMU->rdH(0x04000202)&~2);
	    CPU->interrupt(CPU_INTERRUPT_NORMAL);
	}
    }

    MMU->event(MMU_EVENT_IMMEDIATE, NULL);
    
    // Tell the GUI to stop hblank in 272 cycles
    GUI->eventPush(272, EVENT_HBLANK_END, (vfptr)hblankend, this);
}

// Finish hblank
void gbaGPU::HBLend()
{
    curline++;

    // If that end of hblank took us to line 160, we've come to the vblank.
    // Fire that off to the CPU
    if(curline==160)
    {
	reg[REG_DISPSTAT].b[0] |= 1;
	if(reg[REG_DISPSTAT].b[0] & 8)
	{
	    if((MMU->rdH(0x04000200) & 1) && MMU->rdH(0x04000208))
	    {
		MMU->wrH(0x04000202, MMU->rdH(0x04000202)&~1);
		CPU->interrupt(CPU_INTERRUPT_NORMAL);
	    }
	}
    }
    if(curline>227) curline = 0;

    // Update line counter. If we hit a requested vcount, fire
    // an interrupt.
    reg[REG_VCOUNT].b[0] = curline;
    if(curline == reg[REG_DISPSTAT].b[1])
    {
	reg[REG_DISPSTAT].b[0] |= 4;
	if(reg[REG_DISPSTAT].b[0] & 32)
	{
	    if((MMU->rdH(0x04000200) & 4) && MMU->rdH(0x04000208))
	    {
		MMU->wrH(0x04000202, MMU->rdH(0x04000202)&~4);
		CPU->interrupt(CPU_INTERRUPT_NORMAL);
	    }
	}
    }
    else reg[REG_DISPSTAT].b[0] &= ~4;
    
    REG(DISPSTAT)&=(~2);
    
    // Update the background rotation characteristics. If we're at
    // the top of the screen, reload them.
    BGX[0]+=(s16)REG(BG2PB); BGY[0]+=(s16)REG(BG2PD);
    BGX[1]+=(s16)REG(BG3PB); BGY[1]+=(s16)REG(BG3PD);
    switch(curline)
    {
        case 0:
           REG(DISPSTAT)&=(~1);
           BGX[0]=REG(BG2XL)+(REG(BG2XH)<<16);
           BGX[1]=REG(BG3XL)+(REG(BG3XH)<<16);
           BGY[0]=REG(BG2YL)+(REG(BG2YH)<<16);
           BGY[1]=REG(BG3YL)+(REG(BG3YH)<<16);
           break;
        case 160:
	   REG(DISPSTAT)|=1;
	   break;
    }

    MMU->event(MMU_EVENT_IMMEDIATE, NULL);
    
    // Tell the GUI to draw a line in 960 cycles' time.
    GUI->eventPush(960, EVENT_HBLANK_START, (vfptr)drawline, this);
}

// I/O functions, registered with the MMU at startup.
// Since the GPU hooks to a lot of regions, each function must switch on
// the high bits of the address to work out what's being done.

// NOTE: Byte-wise access to the palette and OAM is prohibited.
u8 gbaGPU::rdB(u32 a)
{
    char str[128];
    switch(a&0x0F000000)
    {
	case 0x04000000:
//            sprintf(str, "Video IO: rdB %08X", a);
//	    Logger::log(pluginName) << str;
	    if(reg[(a&0xFF)>>1].flags & MMU_ALLOW_RD)
	        return reg[(a&0xFF)>>1].b[a&1];
	
        case 0x06000000:
	    if((a&0x00018000) != 0x00018000)
                return VRAMb[a&0x0001FFFF];
	    
	default:
            sprintf(str, "Bad video access: rdB %08X", a);
            Logger::log(pluginName) << str;
            return 0xFF;
    }
}

u16 gbaGPU::rdH(u32 a)
{
//    char str[128];
    u16 out=0xFFFF;
    switch(a&0x0F000000)
    {
	case 0x04000000:
//            sprintf(str, "Video IO: rdH %08X", a);
//	    Logger::log(pluginName) << str;
	    if(reg[(a&0xFF)>>1].flags & MMU_ALLOW_RD)
	        out=reg[(a&0xFF)>>1].data;
	    return out;
        case 0x05000000:
	    return PALh[(a&0x000003FF)>>1];
	case 0x06000000:
	    if((a&0x00018000) != 0x00018000)
	    {
                return VRAMh[(a&0x0001FFFF)>>1];
	    }
	case 0x07000000:
	    return OAMh[(a&0x000003FF)>>1];
	default:
	    return 0;
    }
}

u32 gbaGPU::rdW(u32 a)
{
//    char str[128];
    u32 out=0xFFFFFFFF;
    switch(a&0x0F000000)
    {
	case 0x04000000:
//            sprintf(str, "Video IO: rdW %08X", a);
//	    Logger::log(pluginName) << str;
	    if(reg[(a&0xFF)>>1].flags & MMU_ALLOW_RD)
	        out=reg[(a&0xFF)>>1].data;
	    if(reg[((a&0xFF)>>1)+1].flags & MMU_ALLOW_RD)
	        out|=(reg[((a&0xFF)>>1)+1].data) << 16;
	    return out;
        case 0x05000000:
	    return PALw[(a&0x000003FF)>>2];
	case 0x06000000:
	    if((a&0x00018000) != 0x00018000)
	    {
                return VRAMw[(a&0x0001FFFF)>>2];
	    }
	case 0x07000000:
	    return OAMw[(a&0x000003FF)>>2];
	default:
	    return 0;
    }
}

void gbaGPU::wrB(u32 a, u8  d)
{
    char str[128];
    switch(a&0x0F000000)
    {
	case 0x04000000:
//            sprintf(str, "Video IO: wrB %08X, %02X", a, d);
//	    Logger::log(pluginName) << str;
	    if(reg[(a&0xFF)>>1].flags & MMU_ALLOW_WR)
	        reg[(a&0xFF)>>1].b[a&1]=d;
	    break;
	
        case 0x06000000:
	    if((a&0x00018000) != 0x00018000)
	        VRAMb[a&0x0001FFFF]=d;
	    break;
	    
	default:
            sprintf(str, "Bad video access: wrB %08X, %02X", a, d);
            Logger::log(pluginName) << str;
    }
}

void gbaGPU::wrH(u32 a, u16 d)
{
//    char str[128];
    switch(a&0x0F000000)
    {
	case 0x04000000:
//            sprintf(str, "Video IO: wrH %08X, %04X", a, d);
//	    Logger::log(pluginName) << str;
	    if(reg[(a&0xFF)>>1].flags & MMU_ALLOW_WR)
	        reg[(a&0xFF)>>1].data=d;
	    break;
	case 0x05000000:
	    PALh[(a&0x000003FF)>>1]=d;
	    break;
        case 0x06000000:
	    if((a&0x00018000) != 0x00018000)
	    {
	        VRAMh[(a&0x0001FFFF)>>1]=d;
//		char str[128];
//		sprintf(str, "%08X=%04X", a, d);
//		Logger::log(pluginName) << "VRAM write: " << str;
	    }
	    break;
	case 0x07000000:
	    OAMh[(a&0x000003FF)>>1]=d;
	    break;
    }
}

void gbaGPU::wrW(u32 a, u32 d)
{
//    char str[128];
    switch(a&0x0F000000)
    {
	case 0x04000000:
//            sprintf(str, "Video IO: wrW %08X, %08X", a, d);
//	    Logger::log(pluginName) << str;
	    if(reg[(a&0xFF)>>1].flags & MMU_ALLOW_WR)
	        reg[(a&0xFF)>>1].data=d&65535;
	    if(reg[((a+2)&0xFF)>>1].flags & MMU_ALLOW_WR)
	        reg[((a+2)&0xFF)>>1].data=d>>16;
//	    sprintf(str, "Results: %02X=%04X, %02X=%04X",
//			 (a&0xFF)>>1, reg[(a&0xFF)>>1].data,
//			 ((a+2)&0xFF)>>1, reg[((a+2)&0xFF)>>1].data);
//	    Logger::log(pluginName) << str;
	    break;
	case 0x05000000:
	    PALw[(a&0x000003FF)>>2]=d;
	    break;
        case 0x06000000:
	    if((a&0x00018000) != 0x00018000)
	    {
	        VRAMw[(a&0x0001FFFF)>>2]=d;
//		char str[128];
//		sprintf(str, "%08X=%08X", a, d);
//		Logger::log(pluginName) << "VRAM write: " << str;
	    }
	    break;
	case 0x07000000:
	    OAMw[(a&0x000003FF)>>2]=d;
	    break;
    }
}

// Wrapper functions that will be called by the GUI to service events.
void gbaGPU::hblank(Plugin *in) { ((gbaGPU*)in)->HBLstart(); }
void gbaGPU::hblankend(Plugin *in) { ((gbaGPU*)in)->HBLend(); }
void gbaGPU::drawline(Plugin *in) { ((gbaGPU*)in)->line(); }

// Initialise plugin: Allocate memories, register with MMU
gbaGPU::gbaGPU(std::string name, REQPTR req, UNREQPTR unreq)
{
    pName = std::string(name);
    pClass = pName.substr(0, pName.find(".")+1);
    pRequest = req;
    pUnrequest = unreq;

    pluginName = std::string(pName);

    GUI = (GUIPlugin*)pRequest("UI");
    palbuffer = new u32[264*136];
    palwinID = GUI->subwinCreate(264,136, "Palettes", palbuffer);

    // Allocate VRAMs
    VRAMh = new u16[48*1024];
    PALh  = new u16[512];
    OAMh = new u16[512];
    
    if(!VRAMh || !PALh || !OAMh)
        throw Exception(ERR_GPU_INIT, pName,
                        "Allocation of VRAMs failed.");
    
    VRAMw=(u32*)VRAMh; PALw=(u32*)PALh; OAMw=(u32*)OAMh;
    VRAMb=(u8*)VRAMh; OAMb=(u8*)OAMh;
    
    CPU = (CPUPlugin*)pRequest("GBA_CPU0.main");
    MMU = (MMU32Plugin*)pRequest("GBA_CPU0.mmu");

    for(int i=0; i<16; ++i)
    {
        MMU->rangeReg(0x50+i, rdB, rdH, rdW, wrB, wrH, wrW);
        MMU->rangeReg(0x60+i, rdB, rdH, rdW, wrB, wrH, wrW);
        MMU->rangeReg(0x70+i, rdB, rdH, rdW, wrB, wrH, wrW);
    }
    
    for(int i=0; i<0x30; ++i)
    {
	reg[i].data = 0;
        reg[i].flags = rflags[i];
    }
    
    for(int i=0; i<6; ++i)
        MMU->mmioReg(0x00+i, rdB, rdH, rdW, wrB, wrH, wrW);
    
    Logger::log(pName) << "Registered with MMU.";

    curline = 0;
    Logger::log(pName) << "Initialised.";
}

// Shut down plugin: Free up memories
gbaGPU::~gbaGPU()
{
    if(VRAMh) { delete VRAMh; VRAMb=NULL; VRAMh=NULL; VRAMw=NULL; }
    if(PALh)  { delete PALh;  PALh=NULL;  PALh=NULL;  }
    if(OAMh)  { delete OAMh;  OAMb=NULL;  OAMh=NULL;  OAMh=NULL;  }
    if(palbuffer) { delete palbuffer; palbuffer=NULL; }
   
    pUnrequest("GBA_CPU0.main", 1);
    pUnrequest("GBA_CPU0.mmu", 1);
    pUnrequest("UI", 0); GUI = NULL;
    
    Logger::log(pName) << "Shutdown.";
}

// Reset plugin: Clear memories, read config file
void gbaGPU::reset()
{
    memset(VRAMh, 0, 96*1024);
    memset(PALh,  0, 1024);
    memset(OAMh,  0, 1024);
    memset(palbuffer, 80, 264*136*4);
    
    // Set the background force constants from INI file
    Config ini; ini.read("gbagpu.ini");
    ForceBG[0]=0;
    if(ini.exists("ForceBG0") && ini["ForceBG0"]=="ON")  ForceBG[0]=1;
    if(ini.exists("ForceBG0") && ini["ForceBG0"]=="OFF") ForceBG[0]=2;
    ForceBG[1]=0;
    if(ini.exists("ForceBG1") && ini["ForceBG1"]=="ON")  ForceBG[1]=1;
    if(ini.exists("ForceBG1") && ini["ForceBG1"]=="OFF") ForceBG[1]=2;
    ForceBG[2]=0;
    if(ini.exists("ForceBG2") && ini["ForceBG2"]=="ON")  ForceBG[2]=1;
    if(ini.exists("ForceBG2") && ini["ForceBG2"]=="OFF") ForceBG[2]=2;
    ForceBG[3]=0;
    if(ini.exists("ForceBG3") && ini["ForceBG3"]=="ON")  ForceBG[3]=1;
    if(ini.exists("ForceBG3") && ini["ForceBG3"]=="OFF") ForceBG[3]=2;

    status(0,0);
    
    for(int i=0;i<0x30;++i) reg[i].data = 0;
    
    BGX[0]=0; BGY[0]=0;
    BGX[1]=0; BGY[1]=0;
    REG(BG2PA)=256; REG(BG2PD)=256;
    REG(BG3PA)=256; REG(BG3PD)=256;

    for(int i=0; i<4; ++i) mospy[i]=0;

    // First scanline call: In 960 cycles, come back and draw a line
    GUI->eventPush(960, EVENT_HBLANK_START, (vfptr)drawline, this);
    
    Logger::log(pName) << "Reset";
}

// Provide status: Fill palette-window framebuffer
void gbaGPU::status(int opt1=0, int opt2=0)
{
    Font5x7 prn(palbuffer, 264, 136);

    prn.print("BGpal", 0,0, 0x00FFFFFF);
    prn.print("OBJpal", 136,0, 0x00FFFFFF);
    for(int a=0;a<128;a++)
    {
	for(int b=0;b<128;b++)
	{
	    COLSHIFT((u8*)palbuffer+((a+8)*264+b)*4, PALh[((a&0x78)<<1)+(b>>3)]);
	    COLSHIFT((u8*)palbuffer+((a+8)*264+b+136)*4, PALh[256+((a&0x78)<<1)+(b>>3)]);
	}
    }

    // Once rendered, refresh the window
    GUI->subwinRefresh(palwinID);
}

//---Plugin architecture support-------------------------------------------
// Retrieve Plugin class from outside
// Parameters: plg   - Address of a pointer to a Plugin class to 'new'
//             name  - FQPN of plugin as listed in INI file
//             req   - Pointer to PluginRequest API function
//             unreq - Pointer to PluginUnrequest API function
EXPORTFUNC void getPlugin(Plugin **plg,
                          std::string name,
                          REQPTR req, UNREQPTR unreq)
{
   *plg = new gbaGPU(name, req, unreq);
}

// Provide plugin version information
PLUGININFO *gbaGPU::getinfo()
{
    return &pInfo;
}

// Release plugin from outside
void gbaGPU::release()
{
    // Delete the Test plugin that was 'new'd in getPlugin.
    delete this;
}

/*** EOF: gbagpu.cpp *****************************************************/

